import numpy as numpy

from .IterativeFitter import IterativeFitter
from .GaussErrorDistribution import GaussErrorDistribution
from .LaplaceErrorDistribution import LaplaceErrorDistribution
from .CauchyErrorDistribution import CauchyErrorDistribution
from .UniformErrorDistribution import UniformErrorDistribution
from .PoissonErrorDistribution import PoissonErrorDistribution
#from .BernoulliErrorDistribution import BernoulliErrorDistribution
from .ExponentialErrorDistribution import ExponentialErrorDistribution
from .ClassicProblem import ClassicProblem

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.8.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON ( Java code )
#  *    2016 - 2021 Do Kester

class MaxLikelihoodFitter( IterativeFitter ):
    """
    Base class with methods common to fitters handling ErrorDistributions.

    Author:      Do Kester.

    Attributes
    ----------
    errdis : None | "gauss" | "laplace" | "cauchy" | "poisson" |
                    "uniform" | "exponential"
        None : Use _ChiSq as function to be minimized
        name : use -logLikelihood as function to be minimized from the named
                errordistribution.
    scale : float
        the (fixed) noise scale
    power : float
        power of errdis (if applicable)

    Raises
    ------
    ConvergenceError    Something went wrong during the convergence if the fit.

    """
    def __init__( self, xdata, model, errdis=None, scale=None, power=2.0,
                    **kwargs ):
        """
        Create a new iterative fitter, providing xdatas and model.

        This is a base class. It collects stuff common to all iterative fitters.
        It does not work by itself.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            the model function to be fitted

        errdis : None | "gauss" | "laplace" | "cauchy" | "poisson" |
                        "uniform" | "exponential"
            None : Use _ChiSq as function to be minimized
            name : use -logLikelihood as function to be minimized from the named
                    errordistribution.
        scale : float
            the (fixed) noise scale of errdis (if applicable)
        power : float (2.0)
            the power of errdis ( if applicable)
        kwargs : dict
            Possibly includes keywords from
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        if scale is not None :
            kwargs["fixedScale"] = scale

        super( MaxLikelihoodFitter, self ).__init__( xdata, model, **kwargs )

        self.errdis = errdis
        self.power = power

    def makeFuncs( self, data, weights=None, index=None, ret=3 ) :
        """
        Make connection to the desired func, gradient and hessian.

        Parameters
        ----------
        data : array_like
            the data to be fitted
        weights : array_like or None
            weights on the data
        index : array_like
            indices of the parameters to be fitted.
        ret : 1 or 2 or 3
            return (func), (func,dfunc) or (func,dfunc,hess)
        """
        if not hasattr( self, "errdis" ) or self.errdis is None :
            landscape = _Chisq( self, data, weights, index=index )
            self.isChisq = True
        else :
            scale = 1.0 if self.fixedScale is None else self.fixedScale
            landscape = _LogL( self, data, weights, index=index,
                    errdis=self.errdis, scale=scale, power=self.power )
            self.isChisq = False

        self.landscape = landscape

        if ret == 1 :
            return ( landscape.func )

        # Gradient (or fprime or sometimes jacobian ... confusing)
        if not hasattr( self, "userGradient" ) or not self.userGradient :
            dfunc = None                        ## nonexistent, False or None
        elif callable( self.userGradient ) :
            dfunc = landscape.userdfunc         ## callable function
        else :
            dfunc = landscape.dfunc             ## default, True

        if ret == 2 :
            return ( landscape.func, dfunc )

        return ( landscape.func, dfunc, landscape.hessian )


    def getScale( self ) :
        """
        Return the stdev of the noise.
        """
        if isinstance( self.landscape, _LogL ) :
            return self.landscape.getScale()

        return super( MaxLikelihoodFitter, self ).getScale()


    def getLogLikelihood( self, autoscale=False, var=1.0 ) :
        if self.isChisq :
            return super( MaxLikelihoodFitter, self ).getLogLikelihood(
                        autoscale=autoscale, var=var )

        if autoscale :
            try :
                self.landscape.hypar[0] = self.getScale()
            except IndexError :
                pass
        return - self.landscape.func( self.model.parameters )


    def normalize( self, normdfdp, normdata, weight=1.0 ) :
        """
        Not Implemented.

        Raises
        ------
        NotImplementedError.
        the method is not implemented for MaxLikelihoodFitters

        """
        raise NotImplementedError( "Not implemented for %s." % str( self ) )


    def testGradient( self, par, at, data, weights=None ):
        """
        returns true if the test fails.

        """

        (func,grad) = self.makeFuncs( data, weights=weights, ret=2 )

        df = grad( par )

        dp = 0.001
        pars = par.copy()
        pars[at] -= dp / 2
        r1 = func( pars )
        pars[at] += dp
        r2 = func( pars )
        grat = numpy.subtract( r2, r1 ) / dp

        print( "At %4d  grad : %10.4f  num : %10.4f"%(at, df[at], grat) )

        return ( abs( grat - df[at] ) > 0.001 )

    def __str__( self ):
        return "MaxLikelihoodFitter"


class _Chisq( object ):
    """
    Internal class to provide a chisq landscape.

    It defines
    func (chisq), dfunc (dchisq/dp) and hess (hessian matrix).

    """

    def __init__( self, outer, data, weights, index=None ):
        """
        Parameters
        ----------
        outer : IterativeFitter
            the outer class of this inner class
        data : array_like
            the data
        weights : array_like
            the weights
        index : array_like
            fit index
        """
        self._outer = outer
        self._data = data
        self._weights = weights
        self._index = index

    def func( self, par ):
        """
        Return the function to be minimized.
        In this case
        .. math::
            \chi^2 = \sum( D_i - F(x_i:p) )^2

        """
        param = self._outer.insertParameters( par, index=self._index )
        return self._outer.chiSquared( self._data, params=param,
                    weights=self._weights )

    def userdfunc( self, par ):
        """
        Return an efficient calculation of the gradient.
        """
        return self._outer.userGradient( self._outer.xdata, par, self._data,
                                  weights=self._weights, index=self._index )

    def dfunc( self, par ):
        """
        Return the gradient of func to the parameters p.
        .. math::
            d\chi^2/dp = -2 \sum( D_i - F_i ) dF_i/dp
        """
        param = self._outer.insertParameters( par, index=self._index )
        res = numpy.subtract( self._outer.model.result( self._outer.xdata, param ), self._data )
        if self._weights is not None:
            res = numpy.multiply( res, self._weights )
        desmat = self._outer.getDesign( params=param, index=self._index )

        return 2 * numpy.inner( desmat.transpose(), res )

    def hessian( self, par ):
        """
        Return the Hessian matrix.
        """
        return self._outer.getHessian( params=par, weights=self._weights, index=self._index )

class _LogL( _Chisq ) :
    """
    Internal class to provide a log Likelihood landscape.

    It defines
    func (-logL), dfunc (dlogL/dp) and hess (hessian matrix).

    """
    def __init__( self, outer, data, weights, index=None, errdis='gauss',
                    scale=1.0, power=2.0 ):
        """
        Parameters
        ----------
        outer : IterativeFitter
            the outer class of this inner class
        data : array_like
            the data
        weights : array_like
            the weights
        index : array_like
            fit index
        errdis : "gauss" | "laplace" | "cauchy" | "poisson" | "gengauss"
            errordistribution to be used to calculate logLikelihood
        scale : float (1.0)
            scale if applicable in errdis
        power : float (2.0)
            power if applicable in errdis
        """

        super( _LogL, self ).__init__( outer, data, weights, index=index )

        self.problem = ClassicProblem( outer.model, xdata=outer.xdata, ydata=data,
            weights=weights )

        if errdis == 'gauss' :
            self.errdis = GaussErrorDistribution( scale=scale )
            self.hypar = [scale]
        elif errdis == 'laplace' :
            self.errdis = LaplaceErrorDistribution( scale=scale )
            self.hypar = [scale]
        elif errdis == 'cauchy' :
            self.errdis = CauchyErrorDistribution( scale=scale )
            self.hypar = [scale]
        elif errdis == 'uniform' :
            self.errdis = UniformErrorDistribution( scale=scale )
            self.hypar = [scale]
        elif errdis == 'poisson' :
            self.errdis = PoissonErrorDistribution( )
            self.hypar = []
#        elif errdis == 'bernoulli' :
#            self.errdis = BernoulliErrorDistribution( )
#            self.hypar = []
        elif errdis == 'exponential' :
            self.errdis = ExponentialErrorDistribution( scale=scale, power=power )
            self.hypar = [scale, power]
        else :
            raise ValueError( "Unknown errordistribution %s."%errdis )

    def getScale( self ) :
        errdis = self.errdis
        return errdis.toSigma( errdis.getScale( self.problem ) )


    def func( self, par ) :
        param = self._outer.insertParameters( par, index=self._index )
        param = numpy.append( param, self.hypar )
        return - self.errdis.logLikelihood( self.problem, param )


    def dfunc( self, par ) :
        param = self._outer.insertParameters( par, index=self._index )
        param = numpy.append( param, self.hypar )

        return - self.errdis.partialLogL( self.problem, param, self._index )

    def hessian( self, par ) :
        param = self._outer.insertParameters( par, index=self._index )
        param = numpy.append( param, self.hypar )

        return - self.errdis.hessianLogL( self.problem, param, self._index )





# run with : python3 -m unittest TestEngine

import unittest
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt
from numpy.testing import assert_array_almost_equal as assertAAE

from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestEngine( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """
    def __init__( self, name ):
        super( TestEngine, self ).__init__( name )


    def initEngine( self ):
        m = GaussModel()

        ep1 = ExponentialPrior( scale=2.0 )
        up = UniformPrior( limits=[-10,10] )
        ep2 = ExponentialPrior()
        m.priors = [ep1, up, ep2]

        xdata = numpy.linspace( 0.0, 10.0, 101 )
        data = numpy.zeros( 101, dtype=float )
        data[30:71] = 1.0
        data[40:61] = 2.0
        data[45:56] = 3.0
        data[48:53] = 4.0
        data[50] = 5.0

        numpy.set_printoptions( precision=3, suppress=True )
        return ( m, xdata, data )

    def testEngine( self ):
        print( "\n   Engine Test 1\n" )
        m, xdata, data = self.initEngine()

        errdis = GaussErrorDistribution( )
        errdis.setLimits( [0.1, 10.0] )

        print( errdis.hyperpar[0].prior )
        print( errdis.hyperpar[0].prior.lowLimit )
        print( errdis.hyperpar[0].prior.highLimit )
        print( errdis.hyperpar[0].prior._umin )
        print( errdis.hyperpar[0].prior._urng )

        fi = [0,1,2,-1]
        problem = ClassicProblem( m, xdata, data )

        Tools.printclass( problem )

        allpars = numpy.append( m.parameters, 1.0 )
        wl = WalkerList( problem, 100, allpars, fi )

        engine = Engine( wl, errdis )

        self.enginetest( engine )

        print( "    make copy of engine" )
        copeng = engine.copy()
        self.enginetest( copeng )


    def enginetest( self, engine ) :
        walkers = engine.walkers
        fi = [0,1,2,-1]
        for kw in range( len( walkers ) ) :
            problem = walkers[kw].problem
            p = engine.unit2Domain( problem, engine.rng.rand( 4 ) )
            logL = engine.errdis.logLikelihood( problem, p )
            engine.setWalker( kw, problem, p, logL, fitIndex=fi )
            engine.reportSuccess()
            engine.reportCall()

        engine.reportReject()
        engine.reportFailed()

        self.assertTrue( engine.maxtrials == 5 )
        self.assertTrue( isinstance( engine.rng, numpy.random.RandomState ) )

        print( engine.walkers[0] )
        print( engine.walkers[0].allpars )

        engine.calculateUnitRange()

        print( "UnitR ", engine.unitRange )

        print( "DomR  ", engine.unit2Domain( problem, engine.unitRange ) )
        engine.printReport()



    def suite( cls ):
        return unittest.TestCase.suite( TestEngine.__class__ )


if __name__ == '__main__':
    unittest.main( )





import numpy
from .Model import Model
from .Model import Brackets
from . import Tools
from .Tools import setAttribute as setatt

_author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2011 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2020 Do Kester

class BracketModel( Brackets ):
    """
    BracketModel provides brackets to a chain of Models.

    Its results are exactly the same as the results of the contained model.

    When the contained model is a compound model (a chain of models), the
    BracketModel make a single unit out of it. It acts as a pair of brackets
    in another chain of models. Since compound models can be joined by operations
    other than addition ( there is also subtract, multiply and divide ) brackets
    are needed to distinguish m1 * ( m2 + m3 ) from m1 * m2 + m3.

    BracketModel is automatically invoked when the Model appended to another model,
    is actually a chain of models.

    Model.Brackets is an internal class inside Model.

    Attributes
    ----------
    model : Model
        to be put inside of brackets
    deep : int
        container depth (only for nice printing).

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    Examples
    --------
    Explicit use of BrackeModel

    >>> m1 = GaussModel( )
    >>> m1 += PolynomialModel( 0 )              # Gauss on a constant background
    >>> m2 = BracketModel( m1 )
    >>> m3 = SineModel( )
    >>> m3 *= m2                                # sine * ( gauss + const )
    >>> print( m3 )

    Implicit use of BrackeModel, automatically invoked when m2 is a chain

    >>> m1 = GaussModel( )
    >>> m1 += PolynomialModel( 0 )              # m1 is a chain of models
    >>> m3 = SineModel( )
    >>> m3 *= m1                                # sine * ( gauss + const )
    >>> print( m3 )                             # exactly the same


    Warning
    -------
    BracketModel is about rather advanced model building.

    Notes
    -----
    1. You have to complete the BracketModel, including parameter reduction,
       BEFORE you put it into a model chain.
    2. If you change a BracketModel which is part of a model chain, unexpected result
       might happen.


    """
    #  *************************************************************************
    def __init__( self, model, copy=None, fixed=None, **kwargs ):
        """
        BracketModel

        When constructing a BracketModel existing attributes are lost, except
        parameters that were 'fixed' in the constituent Models. They stay fixed.

        Parameters
        ----------
        model : Model
            to be put in the container.
        copy : BracketModel
            model to be copied
        fixed : dict
            if fixed is not None raise AttributeError
            Use fixed on the constituent models.

        Raises
        ------
        AttributeErrr : When fixed is not None

        """
        if fixed is not None :
            raise AttributeError( "BracketModel cannot have fixed parameters" )

        super( BracketModel, self ).__init__( model, copy=copy, **kwargs )

    def copy( self ):
        """
        Copy a Bracket Model.

        """
        return BracketModel( self.model.copy(), copy=self )





import unittest
import os
import numpy as np
import math
import matplotlib.mlab as mlab
import matplotlib.pyplot as plt

from BayesicFitting import *

from BayesicFitting import formatter as fmt

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2020 Do Kester


class TestPrior2( unittest.TestCase  ) :

    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )
        np.random.seed( 123456 )

    def testPlotUniform( self ) :
        # example data
        NP = 10000
        u = np.random.rand( NP )
        pr = UniformPrior( limits=[-10,10] )
        d = [pr.unit2Domain( v ) for v in u]

        fac = NP / 4
        x = np.asarray( [-12,-10,-10,10,10,12], dtype=float )
        a = fac / pr._range
        y = np.asarray( [0,0,a,a,0,0], dtype=float )
        self.histo( d, pr, fun=(x,y) )

    def testPlotJeffreys( self ) :
        # example data
        NP = 10000
        u = np.random.rand( NP )
        pr = JeffreysPrior( limits=[1.0,9.0] )
        d = [pr.unit2Domain( v ) for v in u]

        fac = NP / 7
        x = np.arange( 51 ) / 5
        y = np.zeros( 51, dtype=float )
        y[5:46] = fac / ( ( math.log( pr.highLimit ) - math.log( pr.lowLimit ) ) * x[5:46] )
        self.histo( d, pr, fun=(x,y) )

    def testPlotExponential( self ) :
        # example data
        NP = 10000
        u = np.random.rand( NP )
        pr = ExponentialPrior( scale=3 )
        d = [pr.unit2Domain( v ) for v in u]

        fac = NP / 3
        x = np.arange( 51 ) * pr.scale / 5
        y = fac * np.exp( - x / pr.scale ) / pr.scale
        self.histo( d, pr, fun=(x,y) )

    def testPlotLaplace( self ) :
        # example data
        NP = 10000
        u = np.random.rand( NP )
        pr = LaplacePrior( scale=4 )
        d = [pr.unit2Domain( v ) for v in u]

        Tools.printclass( pr )

        fac = math.sqrt( 2.0 ) * NP
        fac = NP
        x = ( np.arange( 101 ) - 50 ) * pr.scale / 5
        y = fac * np.exp( - np.abs( x ) / pr.scale ) / ( 2 * pr.scale )
        self.histo( d, pr, fun=(x,y) )

    def testPlotLaplace2( self ) :
        # example data
        NP = 10000
        u = np.random.rand( NP )
        pr = LaplacePrior( scale=4, limits=[-10,None] )
        d = [pr.unit2Domain( v ) for v in u]

        Tools.printclass( pr )

        fac = 0.7 * NP
        x = ( np.arange( 101 ) - 50 ) * pr.scale / 5
        y = fac * np.exp( - np.abs( x ) / pr.scale ) / ( 2 * pr.scale )
        self.histo( d, pr, fun=(x,y) )

    def testPlotGauss( self ) :
        # example data
        NP = 10000
        u = np.random.rand( NP )
        pr = GaussPrior( scale=4 )
        d = [pr.unit2Domain( v ) for v in u]
        x = ( np.arange( 101 ) - 50 ) * pr.scale / 5
        fac = NP / 4
        y = [fac * pr.result( v )/4 for v in x]
        self.histo( d, pr, fun=(x,y) )

    def testPlotCauchy( self ) :
        # example data
        NP = 10000
        u = np.random.rand( NP )
        pr = CauchyPrior( scale=4 )
        d = np.asarray( [pr.unit2Domain( v ) for v in u] )
        # There are too many far outliers in a Cauchy
        q = np.where( d < -50 )
        d[q] = -50
        q = np.where( d > 50 )
        d[q] = 50
        x = ( np.arange( 101 ) - 50 ) * pr.scale / 5
        fac = 1.2 * NP
        y = [fac * pr.result( v ) for v in x]
        self.histo( d, pr, fun=(x,y) )


    def histo( self, x, pr, fun=None ) :
        print( pr )
        print( fmt( x ) )
        print( fmt( fun[0] ) )
        print( fmt( fun[1] ) )

        if self.doplot :
            num_bins = 80
            # the histogram of the data
            n, bins, patches = plt.hist( x, num_bins, facecolor='green', alpha=0.5 )

            # add a 'best fit' line
            if fun is not None :
                plt.plot( fun[0], fun[1], 'r--')
            plt.xlabel('Error')
            plt.ylabel('Probability')
            plt.title('Histogram of ' + str( pr ) )

            # Tweak spacing to prevent clipping of ylabel
            plt.subplots_adjust(left=0.15)
            plt.show()

if __name__ == '__main__':
    unittest.main( )




import numpy as numpy
import math
from . import Tools
from .Formatter import formatter as fmt

from .Engine import Engine

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2022 Do Kester

class OrderEngine( Engine ):
    """
    The OrderEngine is the base engine for all order problems

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213, verbose=0 ):
        """
        Constructor.

        Parameters
        ----------
        walkers : SampleList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : OrderEngine
            to be copied
        seed : int
            for random number generator

        """
        super( ).__init__( walkers, errdis, copy=copy, seed=seed, verbose=verbose )

    def copy( self ):
        """ Return copy of this.  """
        return OrderEngine( self.walkers, self.errdis, copy=self )

    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood, append=False, iteration=0 ):
        """ 
        Execute the engine by diffusing the parameters.
        
        Parameters
        ----------   
        kw : walker-id
            walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        append : bool
            not used here
        iteration : int
            iteration number        

        Returns
        -------
        int : the number of successfull moves
           
        """
        self.reportCall()
        t = 0
        k = 0
        while t == 0 and k < self.maxtrials :
            t = self.executeOnce( kw, lowLhood )
            k = k + 1
        
        return t

    def __str__( self ):
        return str( "OrderEngine" )

    def calculateUnitRange( self ) :
        """
        Irrelevant for OrderProblems
        """
        pass



import numpy as numpy
from astropy import units
import re
import warnings
from .Problem import Problem
from .Dynamic import Dynamic
from . import Tools

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Alpha"


#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2022  Do Kester

class OrderProblem( Problem ):
    """
    An OrderProblem needs to optimize the order of a set of nodes.
    the nodes are given by the x variable; the order by the parameters p.

    The result of the function for certain x and p is given by
    `problem.result( x, p )`

    This class is a base class. Further specializations will define the
    result method.

    Attributes
    ----------
    parameters : array_like
        to be optimized in TBD ways
    npbase : int
        number of params in the base model
    ndim : int
        number of dimensions of input. (default : 1)

    Author :         Do Kester

    """

    #  *************************************************************************
    def __init__( self, xdata=None, weights=None, copy=None ):
        """
        OrderProblem Constructor.
        <br>
        Parameters
        ----------
        xdata : array_like of shape [np,ndim]
            the nodes to be visited
        weights : array_like
            weights on the arrival nodes
        copy : BaseProblem
            to be copied

        """

        super( ).__init__( xdata=xdata, weights=weights, copy=copy )

        self.npars, self.ndim = self.xdata.shape
        self.ndata = self.npars

        self.partype = int

        if copy is None :
            self.parameters = numpy.arange( self.npars, dtype=int )
        else :
            self.parameters = copy.parameters


    def copy( self ) :
        """ Return a copy. """
        return OrderProblem( copy=self )

    def isDynamic( self ) :
        return isinstance( self, Dynamic )


    def myEngines( self ) :
        myeng = ["move", "reverse", "switch", "loop"]

        if self.isDynamic() :
            return myeng + ["birth", "death"]
        else :
            return myeng

    def myStartEngine( self ) :
        return "startorder"

    def myDistribution( self ) :
        return "distance"

    def baseName( self ) :
        return "OrderProblem"




# run with : python3 -m unittest TestSalesmanProblem

import unittest
import os
import numpy as numpy
from astropy import units
from numpy.testing import assert_array_equal as assertAE
import math
import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import formatter as fmt
from BayesicFitting import fma


__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for solving the SalesmanProblem with NestedSampler.

    Author       Do Kester

    """
    def __init__( self, name ):
        super( ).__init__( name )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def initProblem( self, np=10, random=False ):

        n2 = np * np
        if random :
            numpy.random.seed( 12345 )
            x0 = numpy.random.rand( n2 ) * np
            x1 = numpy.random.rand( n2 ) * np
        else :
            x0 = numpy.arange( n2, dtype=int ) % np
            x1 = numpy.arange( n2, dtype=int ) // np
        xdata = numpy.append( x0, x1 ).reshape( 2, n2 ).transpose()
        xdata = numpy.array( xdata, dtype=float )

        problem = SalesmanProblem( xdata )

#        numpy.set_printoptions( precision=3, suppress=True )
        return problem

    def test1( self ):
        print( "\n   SalesmanProblem Test MoveEngine\n" )
        self.stdenginetest( MoveEngine, np=4, plot=self.doplot, random=True )

    def test2( self ):
        print( "\n   SalesmanProblem Test ReverseEngine\n" )
        self.stdenginetest( ReverseEngine, np=5, plot=self.doplot )

    def test3( self ):
        print( "\n   SalesmanProblem Test ShuffleEngine\n" )
        self.stdenginetest( ShuffleEngine, np=5, plot=self.doplot )

    def test4( self ):
        print( "\n   SalesmanProblem Test SwitchEngine\n" )
        self.stdenginetest( SwitchEngine, np=5, plot=self.doplot )

    def test5( self ):
        print( "\n   SalesmanProblem Test LoopEngine\n" )
        self.stdenginetest( LoopEngine, np=5, plot=self.doplot, random=True )

    def test6( self ):
        print( "\n   SalesmanProblem Test NearEngine\n" )
        self.stdenginetest( NearEngine, np=5, plot=self.doplot, random=True )

    def teststart1( self ) :
        print( "\n   SalesmanProblem Test StartOrderEngine\n" )
        self.stdstarttest( StartOrderEngine, np=4, nwalker=10 )

    def XXXteststart2( self ) :
        print( "\n   SalesmanProblem Test StartNearEngine\n" )
        self.stdstarttest( StartNearEngine, np=4, nwalker=10 )


    def stdstarttest( self, mystarteng, np=4, nwalker=10 ) :

        problem = self.initProblem( np=np )

        errdis = DistanceCostFunction( )

        n2 = np * np
        pars = numpy.arange( n2, dtype=int )
        fi = numpy.arange( n2, dtype=int )
        sl = WalkerList( problem, nwalker, pars, fi )

        steng = mystarteng( sl, errdis, verbose=5 )
        for wlkr in steng.walkers :
            logL = errdis.logLikelihood( problem, wlkr.allpars )
            wid = wlkr.id
            print( "Walker :", wid, wlkr.allpars, fmt( logL ) )
            steng.execute( wid, -math.inf )

            nwlk = steng.walkers[wid]
            print( "          ", nwlk.allpars, fmt( nwlk.logL ) )


            psort = numpy.sort( nwlk.allpars )
            assertAE( psort, pars )


    def stdenginetest( self, myengine, nwalker=4, np=10, plot=False, random=False ) :
        problem = self.initProblem( np=np, random=random )

        errdis = DistanceCostFunction( )
#        print( problem.xdata )

        n2 = np * np
        pars = numpy.arange( n2, dtype=int )
        fi = numpy.arange( n2, dtype=int )
        sl = WalkerList( problem, nwalker, pars, fi )

        if not random :
            n1 = np - 1
            m = n1 * n1
            s = np * n1
            s += math.sqrt( m + 1 ) * ( np - 1 )
            s += math.sqrt( m + m )

            logL = numpy.sum( problem.result( pars ) )
            print( s, logL, abs( s - logL ) )
            self.assertTrue( abs( s - logL ) < 1.0e-10 )

        steng = StartOrderEngine( sl, errdis, verbose=5 )
        myeng = myengine( sl, errdis, verbose=5 )

        for wlkr in steng.walkers :
            print( "=========", wlkr )

            pars = wlkr.allpars
#            print( "Before ", wlkr.allpars )

            wid = wlkr.id
            myeng.execute( wid, -math.inf )

            neww = steng.walkers[wid]
#            print( "After  ", neww.allpars )
            newp = neww.allpars

            psort = numpy.sort( neww.allpars )
            assertAE( psort, pars )

            if plot :
                xy0 = problem.xdata[pars]
                plt.plot( xy0[:,0], xy0[:,1], 'k-' )
                plt.plot( [xy0[-1,0], xy0[0,0]], [xy0[-1,1], xy0[0,1]], 'b-' )
                xy1 = problem.xdata[newp]
                plt.plot( xy1[:,0], xy1[:,1], 'r-.' )
                plt.plot( [xy1[-1,0], xy1[0,0]], [xy1[-1,1], xy1[0,1]], 'g-.' )
                for kp in range( n2 ) :
                    plt.text( xy0[kp,0], xy0[kp,1], "%d"%kp,
                               {'color': 'black', 'ha': 'right'} )
                    plt.text( xy1[kp,0], xy1[kp,1], "%d"%kp,
                               {'color': 'red', 'ha': 'left'} )


                plt.show()

        myeng.verbose = 0
        wid = wlkr.id
        for k in range( 1000 ) :
            myeng.execute( wid, -math.inf )
            psort = numpy.sort( steng.walkers[wid].allpars )
            assertAE( psort, pars )

    def suite( cls ):
        return unittest.TestCase.suite( TestOrderEngines.__class__ )


if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
from astropy import units
import math
from . import Tools

from .IterativeFitter import IterativeFitter
from .ConvergenceError import ConvergenceError
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2020 Do Kester


class LevenbergMarquardtFitter( IterativeFitter ):
    """
    Non-linear fitter using the Levenberg-Marquardt method.

    Implementation of the Levenberg-Marquardt algorithm to fit the parameters
    of a non-linear model. It is a gradient fitter which uses partial
    derivatives to find the downhill gradient. Consequently it ends in the
    first minimum it finds.

    The original C-version stems from Numerical Recipes with some additions of my own.
    This might be the third or fourth transcription of it.

    Author       Do Kester.

    Examples
    --------
    # assume x and y are Double1d data arrays:
    >>> x = numpy.arange( 100, dtype=float ) / 10
    >>> y = numpy.arange( 100, dtype=float ) / 122            # make slope
    >>> rg = RandomGauss( seed=12345L )            # Gaussian random number generator
    >>> y += rg( numpy.asarray( 100, dtype=float ) ) * 0.2            # add noise
    >>> y[Range( 9,12 )] += numpy.asarray( [5,10,7], dtype=float )         # make some peak
    # define a model: GaussModel + background polynomial
    >>> gauss = GaussModel( )                            # Gaussian
    >>> gauss += PolynomialModel( 1 )                    # add linear background
    >>> gauss.setParameters( numpy.asarray( [1,1,0.1,0,0], dtype=float ) )    # initial parameter guess
    >>> print gauss.getNumberOfParameters( )                # 5 ( = 3 for Gauss + 2 for line )
    >>> gauss.keepFixed( numpy.asarray( [2] ), numpy.asarray( [0.1], dtype=float ) )    # keep width fixed at 0.1
    >>> lmfit = LevenbergMarquardtFitter( x, gauss )
    >>> param = lmfit.fit( y )
    >>> print param.length( )                             # 4 ( = 5 - 1 fixed )
    >>> stdev = lmfit.getStandardDeviation( )             # stdevs on the parameters
    >>> chisq = lmfit.getChiSquared( )
    >>> scale = lmfit.getScale( )                         # noise scale
    >>> yfit  = lmfit.getResult( )                        # fitted values
    >>> yband = lmfit.monteCarloError( )                       # 1 sigma confidence region
    # for diagnostics ( or just for fun )
    >>> lmfit = LevenbergMarquardtFitter( x, gauss )
    >>> lmfit.setVerbose( 2 )                             # report every 100th iteration
    >>> plotter = IterationPlotter( )                     # from BayesicFitting
    >>> lmfit.setPlotter( plotter, 20 )                   # make a plot every 20th iteration
    >>> param = lmfit.fit( y )

    Notes
    -----
    In case of problems look at the "Troubles" page in the documentation area.


    Limitations
    -----------
    1. LMF is <b>not</b> guaranteed to find the global minimum.
    2. The calculation of the evidence is an Gaussian approximation which is
    only exact for linear models with a fixed scale.

    Attributes
    ----------
    xdata : array_like
        vector of numbers as input for model
    model : Model
        the model to be fitted
    lamda : float
        to balance the curvature matrix (see Numerical Recipes)

    """
    #  *************************************************************************
    def __init__( self, xdata, model, **kwargs ):
        """
        Create a class, providing xdata and model.

        Parameters
        ----------
        xdata : array_like
            vector of independent input values
        model : Model
            a model function to be fitted

        kwargs : dict
            Possibly includes keywords from
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale


        """
        super( LevenbergMarquardtFitter, self ).__init__( xdata, model, **kwargs )

        self.lamda = 0.001
        self.converged = False
        self.first = True

    #  *************************************************************************
    def fit( self, data, weights=None, par0=None, keep=None, limits=None,
                maxiter=None, tolerance=None, verbose=None, plot=False,
                callback=None ):
        """
        Return Model fitted to the data arrays.

        It will calculate the hessian matrix and chisq.

        Parameters
        ----------
        data  : array_like
            the data vector to be fitted
        weights : array_like
            weights pertaining to the data
        par0 : array_like
            initial values for the parameters of the model
            default: from model
        keep : dict of {int : float}
            dictionary of indices (int) of parameters to be kept at fixed value (float)
            The values of `keep` are only valid for *this* fit
            see also `LevenbergMarquardtFitter( ..., keep=dict )
        limits : None or list of 2 floats or list of 2 array_like
            None : no limits applied
            [lo,hi] : low and high limits for all values of the parameters
            [la,ha] :  arrays of low and high limits for all values of the parameters
        maxiter : int
            max number of iterations. default=1000,
        tolerance : float
            absolute and relative tolrance. default=0.0001,
        verbose : int
            0 : silent
            >0 : more output
            default=1
        plot : bool
            plot the results
        callback : callable
            is called each iteration as
            `val = callback( val )`
            where val is the parameter list.

        Raises
        ------
        ConvergenceError if it stops when the tolerance has not yet been reached.

        """
        if maxiter is None : maxiter = self.maxIter
        if tolerance is None : tolerance = self.tolerance
        if verbose is None : verbose = self.verbose

        fitIndex, data, weights = self.fitprolog( data, weights=weights, keep=keep )


        trypar = self.model.parameters if par0 is None else par0

#        if fitIndex is not None and len( fitIndex ) < len( par0 ) :
#            par0 = par0[fitIndex]

        self.chi = self.chiSquaredExtra( data, trypar, weights=weights ) + 1

        self.lamda = 0.001

        if verbose > 1 and self.first :
            print( "  Iter  loglambda    chisq   parameters" )
            self.first = False

        while self.iter < maxiter :

            trypar, trychi = self.trialfit( trypar, fitIndex, data, weights, verbose, maxiter )
            self.model.parameters = trypar

            tol = tolerance if self.chi < 1 else tolerance * self.chi
            if abs( self.chi - trychi ) < tol :
                self.chi = trychi
                self.report( verbose, trypar, trychi, more=math.log10( self.lamda ), force=True )
                self.converged = True

                self.fitpostscript( data, plot=plot )
                return trypar                   # we are done.

            self.chi = trychi
            if self.lamda > 1e-100 :
                self.lamda *= 0.1

        raise ConvergenceError( "LevenbergMarquardtFitter.fit( ): " +
                                "Too many iterations: ", self.iter )


    def chiSquaredExtra( self, data, params, weights=None ) :
        """
        Add normalizing data to chisq.
        """
        chisq = self.chiSquared( data, params=params, weights=weights )
        if hasattr( self, "normdfdp" ) :
            res = numpy.inner( self.normdfdp, params ) - self.normdata
            res *= res * self.normweight
            chisq += numpy.sum( res )

        return chisq


    #  *************************************************************************
    def trialfit( self, params, fi, data, weights, verbose, maxiter ):

        hessian = self.getHessian( params=params, weights=weights, index=fi )

        residu = data - self.model.result( self.xdata, params )
        if weights is not None :
            residu *= weights
        if hasattr( self, "normdfdp" ) :
            nres = ( self.normdata - numpy.inner( self.normdfdp, params ) ) * self.normweight
            residu = numpy.append( residu, nres )

        vector = self.getVector( residu, index=fi )

        nfit = len( fi )
        fitpar = params[fi]

        while self.iter < maxiter :

            for k in range( nfit ) :
                hessian[k,k] *= ( 1 + self.lamda )

            newpar = fitpar + 0.5 * numpy.linalg.solve( hessian, vector )

            onEdge, edgePar, edgeInd = self.checkLimits( newpar, fi )

            if onEdge :
                newpar = edgePar

            trypar = self.insertParameters( newpar, index=fi, into=params )

            trychi = self.chiSquaredExtra( data, trypar, weights=weights )

            self.ntrans += 1
            self.iter += 1

#            print( "trialfit", self.chi, trychi, math.log10( self.lamda ) )
            self.report( verbose, trypar, trychi, more=math.log10( self.lamda ),
                         force=(verbose >= 3) )

            if trychi <= self.chi:

                if onEdge :                        # further convergence on the edge plane(s)
                    self.chi = trychi
                    self.model.parameters = trypar
                    trypar, trychi = self.trialfit( trypar, edgeInd, data, weights,
                            verbose, maxiter )


                return ( trypar, trychi )          #  succesfull step

            if self.lamda < 1e20 :
                self.lamda *= 10
            self.fitpar = trypar                   #  keep to report back

        raise ConvergenceError( "LevenbergMarquardtFitter. Too many iterations: ", self.iter )


    ### TBD to __getattr__  ???
    def getParameters( self ):
        """
        Return status of the fitter: parameters ( for debugging ).

        Only for debugging; use Model.getParameters( ) otherwise

        """
        return self.model.parameters if self.converged else self.fitpar

    def __str__( self ):
        """ Return name of the fitter.  """
        return "LevenbergMarquardtFitter"

    def checkLimits(  self, fitpar, fitindex ):
        """

        """
        onedge = False
#        if self.model.priors is not None and self.model.priors.hasLimits( ):
        if self.model.priors is not None :
            fitin = []
            for i,k in enumerate( fitindex ) :
                pr = self.model.getPrior( k )
                if pr.hasLowLimit( ) and fitpar[i] < pr.lowLimit :
                    fitpar[i] = pr.lowLimit
                    onedge = True
                elif pr.hasHighLimit( ) and fitpar[i] > pr.highLimit :
                    fitpar[i] = pr.highLimit
                    onedge = True
                else:
                    fitin += [k]

            return ( onedge, fitpar, fitin )
        else:
            return ( onedge, fitpar, fitindex )







import math

from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.6.2"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2020 Do Kester


class LaplacePrior( Prior ):
    """
    Laplace prior distribution.

        Pr( x ) = exp( - |x - center| / scale )

    By default: center = 0.0 and scale = 1.

    It can also have a limited domain.
    By default the domain is [-Inf,+Inf].
    In computational practice the domain is limited to about [-36,36] scale units

    Equivalent to a double-sided exponential prior

    domain2unit: u = 0.5 * exp( ( d - c ) / scale )             if d < c
                     1.0 - 0.5 * exp( ( c - d ) / scale )       otherwise
    unit2domain: d = c + log( 2 * u ) * scale                   if u < 0.5
                     c - log( 2 * ( 1 - u ) ) * scale           otherwise

    Examples
    --------
    >>> pr = LaplacePrior()                         # center=0, scale=1
    >>> pr = LaplacePrior( center=1.0, scale=0.5 )
    >>> pr = LaplacePrior( limits=[0,None] )        # limites to values >= 0
    >>> pr = LaplacePrior( center=1, circular=3 )   # circular between 0.5 and 2.5

    Attributes
    ----------
    center : float
        center of the Laplace prior
    scale : float
        scale of the Laplace prior

    Attributes from Prior
    --------------------=
    lowLimit, highLimit, deltaP, _lowDomain, _highDomain

    lowLimit and highLimit cannot be used in this implementation.

    """

    MAXVAL = 40         ## scale units

    #  *********CONSTRUCTOR***************************************************
    def __init__( self, center=0.0, scale=1.0, limits=None, circular=False, prior=None ):
        """
        Constructor.

        Parameters
        ----------
        center : float
            of the prior
        scale : float
            of the prior
        limits : None or list of 2 float/None
            None : no limits.
            2 limits, resp low and high
        circular : bool or float
            bool : y|n circular with period from limits[0] to limits[1]
            float :period of circularity
        prior : LaplacePrior
            prior to copy (with new scale if applicable)

        """
        self.center = center
        self.scale = scale

        super( LaplacePrior, self ).__init__( limits=limits, circular=circular, prior=prior )


    def copy( self ):
        return LaplacePrior( prior=self, center=self.center, scale=self.scale,
                             limits=self.limits, circular=self.circular )

    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for a Laplace distribution.

        domain2unit: u = 0.5 * exp( ( d - c ) / s ) if d < c else
                         1.0 - 0.5 * exp( ( c - d ) / s )

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        d = dval - self.center
        uval = ( 0.5 * math.exp(  d / self.scale ) if ( d < 0 ) else
           1.0 - 0.5 * math.exp( -d / self.scale ) )

        return uval

    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for a Laplace distribution.

        unit2domain: d = c + log( 2 * u ) * scale if u < 0.5 else
                         c - log( 2 * ( 1 - u ) ) * scale;

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        scl = self.scale
        if uval > 0.5:
            uval = 1 - uval
            scl *= -1
        try :
            return self.center + math.log( 2 * uval ) * scl
        except :
            return self.center - self.MAXVAL * scl

    def result( self, x ):
        """
        Return a the result of the distribution function at x.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        if self.isOutOfLimits( x ) : return 0
        return 0.5 * math.exp( -abs( x - self.center ) / self.scale ) / self.scale

    def logResult( self, x ):
        """
        Return a the log of the result of the distribution function to p.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        if self.isOutOfLimits( x ) : return -math.inf

        xc = x - self.center
        return math.log( 0.5 / self.scale ) - abs( xc ) / self.scale

    def partialLog( self, x ):
        """
        Return partial derivative of log( Prior ) wrt parameter.

        Parameters
        ----------
        x : float
            the value

        """
        if x == self.center : return 0                  ## actually a cusp
        elif self.isOutOfLimits( x ) : return math.nan

        return -1 / self.scale if x > self.center else 1 / self.scale

    def isBound( self ):
        """ Return true if the integral over the prior is bound.  """
        return True

    def shortName( self ) :
        return "LaplacePrior"





# run with : python3 -m unittest TestAmoebaFitter

import unittest
import os
import numpy as numpy
from numpy.testing import assert_array_almost_equal as assertAAE
from astropy import units
import math

import matplotlib.pyplot as plt


from BayesicFitting import *


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2004 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class TestAmoebaFitter( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    aa = 5              #  average offset
    bb = 2              #  slope
    ss = 0.3            #  noise Normal distr.

    x = numpy.asarray( [ -1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
    noise = numpy.asarray( [-0.000996, -0.046035,  0.013656,  0.418449,  0.0295155,  0.273705,
      -0.204794,  0.275843, -0.415945, -0.373516, -0.158084] )

    def eq( self, a, b, eps=1e-10 ) :
        if ( a + b ) != 0 :
            return abs( a - b ) / abs( a + b ) < eps
        else :
            return abs( a - b ) < eps


################################################################################
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def testAmoebaFitter( self ):

        c0 = 3.2
        c1 = -0.1
        c2 = 0.3
        c3 = 1.1
        c4 = 2.1
        y = ( self.x - c1 ) / c2
        y = c0 * numpy.exp( -y * y ) + self.noise

        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        print( "Testing Nonlinear Fitters: Amoeba" )
        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        modl1 = GaussModel( )

        numpy.set_printoptions(  precision=3, suppress=True )
        amfit = AmoebaFitter( self.x, modl1 )

        tr = numpy.asarray( [c0,c1,c2,c3,c4] )
        print( "Truth ", tr )

        par1 = amfit.fit( y )

        print( "x     ", self.x )
        print( "y     ", y )
        print( "par   ", par1 )

        modl2 = GaussModel( )
        modl2.addModel( PolynomialModel(1) )
        z = y + c4 * self.x + c3

        modl2.parameters = numpy.append( par1, [0,0] )
        lmfit = AmoebaFitter( self.x, modl2 )

        par2 = lmfit.fit( z )
        print( "z      ", z )
        print( "par2   ", par2 )


        print( "chisq1 = ", amfit.chisq, "  chisq2 = ", lmfit.chisq )

        assertAAE( par1, par2[:3], 1 )
        assertAAE( par2, tr, 0 )

        if self.doplot :
            xx = numpy.linspace( -1, +1, 1001 )
            plt.plot( self.x, y, 'k+' )
            plt.plot( xx, modl1.result( xx ), 'k-' )
            plt.plot( self.x, z, 'r+' )
            plt.plot( xx, modl2.result( xx ), 'r-' )
            plt.show()



    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestAmoebaFitter.__class__ )





import numpy as numpy
from astropy import units
import math
from . import Tools

from .HyperParameter import HyperParameter
from .JeffreysPrior import JeffreysPrior

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2011 - 2014 Do Kester, SRON (JAVA code)
#  *    2016 - 2020 Do Kester

class NoiseScale( HyperParameter ):
    """
    Hyperparameter for the scale of a ScaledErrorDistribution

    it is a measure of the noise.

    Information about the scale of the noise is stored in his class.
    It is either in the form of a fixed number, when the noise scale
    is known or in the form of a Prior with limits.
    By default this prior is a JeffreysPrior..

    The full use of priors is reserved for Bayesian calculations as
    in NestedSampler

    Attributes
    ----------
    scale : float
        the value of the noiseScale.  Default: 1.0
    stdev : float
        the standard deviation of the noise scale.  Default: None
    prior : Prior
        the prior for the noiseScale.  Default: JeffreysPrior
    fixed : boolean
        keep the noise scale fixed at the value given by scale.
        default: True
    minimum : boolean
        automatic noise scaling with a minimum. default: False

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, scale=1.0, isFixed=True, prior=None, limits=None,
                    copy=None ):
        """
        Constructor.

        Parameters
        ----------
        scale : float
            float   value of the noise scale
        isFixed : bool
            True:   Consider the hyperparameter as fixed
            False:  Optimize the parameter too (when relevant)
                    It might need a prior and/or limits to be set
                    The default prior is JeffreysPrior
        prior : None or Prior
            None : no prior set
            Prior : the prior probability on scale
        limits : None or list of 2 floats
            None : no limits set
            [lo,hi] : limits to be passed to the Prior.
            If limits are set, the default for Prior is JeffreysPrior
        copy : NoiseScale
            NoiseScale to copy

        """
        if limits is not None and prior is None :
            prior = JeffreysPrior()
        super( NoiseScale, self ).__init__( hypar=scale, isFixed=isFixed,
                        prior=prior, limits=limits )
        self.minimum = False

        if copy is not None :
            self.minimum = copy.minimum

    def copy( self ):
        """ Return a copy.  """
        return NoiseScale( scale=self.scale, copy=self )

    def __setattr__( self, name, value ) :
        """
        Rename scale to hypar and stdevScale to stdev.
        """
        if name == "scale" :
            self.hypar = value
        elif name == "stdevScale" :
            self.stdev = value
        else :
            object.__setattr__( self, name, value )


    def __getattr__( self, name ) :
        if name == "scale" :
            return self.hypar
        elif name == "stdevScale" :
            return self.stdev
        else :
            raise AttributeError( "Unknown attribute " + name )

        return None

    def minimumScale( self, scale=None ) :
        """
        Fit the noise scale with a minimum value.

        Parameters
        ----------
        scale : float
            the value of the noise scale. Default: noiseScale.scale

        """
        if scale is not None : self.hypar = scale
        self.minimum = True

    def __str__( self ) :
        return str( "Noise scale. value = %f" % self.hypar )





## run as : python3 -m unittest TestProblem

from __future__ import print_function

import numpy as numpy
import os
from numpy.testing import assert_array_almost_equal as assertAAE
import unittest
from astropy import units
import math
import sys
import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import formatter as fmt
from BayesicFitting import fma

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2017 Do Kester

class TestProblem( unittest.TestCase ):
    """
    Test harness for Problem classes.

    Author       Do Kester

    """
    SQRT2 = math.sqrt( 2 )
    LGSQ2 = math.log( SQRT2 )
    LOG2 = math.log( 2 )
    EPSILON = 1e-6

    # Define x independent variable
    x = numpy.asarray( [ -1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0], dtype=float )

    # Define y dependent variable
    data = numpy.asarray( [-11.1, -9.2, -5.1, -5.2, -1.1, 1.2, 1.1, 5.2, 5.1, 8.2, 11.2], dtype=float )

    # Define wgt weigth
    wgt = numpy.asarray( [1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=float )

    def __init__( self, name ):
        super( ).__init__( name )
        numpy.set_printoptions(formatter={'float': '{: 0.3f}'.format})
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def test1( self ) :
        print( "====test1 (Classic)Problem============================" )
        nn = 11
        x = numpy.linspace( -1, 1, nn, dtype=float )
        ym = self.x

        model = PolynomialModel( 0 )
        model.parameters = 0.0
        problem = ClassicProblem( model=model, xdata=x, ydata=ym )
#        fitIndex=[0]

        print( problem.shortName() )
        Tools.printclass( problem )

        self.assertTrue( problem.sumweight == problem.ndata )
        self.assertFalse( problem.hasWeights() )
        r = problem.residuals( [0.0] )
        print( fma(r) )
        print( fmt(numpy.sum( r )), fmt(numpy.sum( numpy.abs( r ) )) )
        self.assertAlmostEqual( numpy.sum( r ), 0 )
        self.assertAlmostEqual( numpy.sum( numpy.abs( r ) ), 6 )

        r, s = problem.weightedResiduals( [0.0], extra=True )
        self.assertAlmostEqual( numpy.sum( r ), 0 )
        self.assertAlmostEqual( numpy.sum( numpy.abs( r ) ), 6 )
        self.assertTrue( s[0] == s[4] == -1 )
        self.assertTrue( s[6] == s[10] == 1 )


        r,s = problem.weightedResSq( [0.0], extra=True )
        self.assertAlmostEqual( numpy.sum( r ), 4.4 )
        self.assertAlmostEqual( numpy.sum( numpy.abs( r ) ), 4.4 )
        self.assertTrue( s[0] + s[4] == -1.2 )
        self.assertTrue( s[6] + s[10] == 1.2 )

        problem.weights = self.wgt

        self.assertTrue( problem.sumweight == 16 )
        self.assertTrue( problem.hasWeights() )

        r = problem.residuals( [0.0] )
        self.assertAlmostEqual( numpy.sum( r ), 0 )
        self.assertAlmostEqual( numpy.sum( numpy.abs( r ) ), 6.0 )


        r, s = problem.weightedResiduals( [0.0], extra=True )
        self.assertTrue( numpy.sum( r ) == 0.0 )
        self.assertTrue( numpy.sum( numpy.abs( r ) ) == 8.4 )

        self.assertTrue( s[0] == s[4] == -1 )
        self.assertTrue( s[1] == s[3] == -2 )
        self.assertTrue( s[6] == s[10] == 1 )
        self.assertTrue( s[7] == s[9] == 2 )

        r,s = problem.weightedResSq( [0.0], extra=True )
        self.assertTrue( numpy.sum( r ) == 6.0 )
        self.assertTrue( s[0] + s[4] == -1.2 )
        self.assertTrue( s[6] + s[10] == 1.2 )

        engs = problem.myEngines()
        self.assertTrue( len( engs ) == 2 )
        self.assertTrue( engs[0] == "galilean" )
        self.assertTrue( engs[1] == "chord" )

        self.assertTrue( problem.myStartEngine() == "start" )
        self.assertTrue( problem.myDistribution() == "gauss" )

    def test2( self ) :
        print( "====test2 ErrorsInXandYProblem============================" )
        nn = 4
        x = numpy.asarray( [0,2,8,10],dtype=float )
        y = numpy.asarray( [2,0,10,8],dtype=float )


        model = PolynomialModel( 1 )
        par = [0.0, 1.0] + [1.0, 1.0, 9.0, 9.0]
        npar = len( par )

        problem = ErrorsInXandYProblem( model=model, xdata=x, ydata=y )

        print( problem.shortName() )
        Tools.printclass( problem )

        self.assertTrue( problem.sumweight == problem.ndata )
        self.assertFalse( problem.hasWeights() )
        md = problem.result( par )
        print( "mock  ", fma( md ) )
        r = problem.residuals( par, mockdata=md )
        print( "res   ", fma(r) )
        print( "sum   ", fmt(numpy.sum( r )), fmt(numpy.sum( numpy.abs( r ) )) )
        sqrt2 = math.sqrt( 2 )
        self.assertAlmostEqual( numpy.sum( r ), 4*sqrt2 )
        self.assertAlmostEqual( numpy.sum( numpy.abs( r ) ), 4*sqrt2 )

        r, s = problem.weightedResiduals( par, extra=True )
        print( "res   ", fma( r ) )
        print( "      ", fma( s ) )
        self.assertAlmostEqual( numpy.sum( r ), 4*sqrt2 )
        self.assertAlmostEqual( numpy.sum( s ), 0 )
        for k in [1,3,4,6] :
            self.assertAlmostEqual( s[k], -sqrt2/2 )
        for k in [0,2,5,7] :
            self.assertAlmostEqual( s[k],  sqrt2/2 )


        r,s = problem.weightedResSq( par, extra=True )
        print( "res2  ", fma( r ) )
        print( "      ", fma( s ) )
        self.assertAlmostEqual( numpy.sum( r ), 8.0 )
        self.assertAlmostEqual( numpy.sum( s ), 0.0 )

        self.assertAlmostEqual( s[0] + s[2] + s[5] + s[7],  4 )
        self.assertAlmostEqual( s[1] + s[3] + s[4] + s[6], -4 )

        problem.weights = numpy.asarray( [1,1,2,2], dtype=float )

        self.assertTrue( problem.sumweight == 6 )
        self.assertTrue( problem.hasWeights() )

        r,s = problem.weightedResSq( par, extra=True )
        print( "res2w ", fma( r ) )
        print( "      ", fma( s ) )
        self.assertTrue( numpy.sum( r ) == 12.0 )
        self.assertTrue( numpy.sum( s ) == 0 )

        engs = problem.myEngines()
        self.assertTrue( len( engs ) == 3 )
        self.assertTrue( engs[0] == "galilean" )
        self.assertTrue( engs[1] == "gibbs" )
        self.assertTrue( engs[2] == "chord" )

        self.assertTrue( problem.myStartEngine() == "start" )
        self.assertTrue( problem.myDistribution() == "gauss" )

        print( "=======   Try Gauss Error Distribution  ==================" )

        ## reset the weights
#        problem.weights = None

        ged = GaussErrorDistribution( )
        param = numpy.append( par, [1.0] )

        fitIndex = [k for k in range( npar )] + [-1]

        lLdata = ged.logLdata( problem, param )
        print( "ldata ", fma( lLdata ) )

        logL0 = numpy.sum( lLdata )

        logL = ged.logLikelihood( problem, param )
        altL = ged.logLikelihood_alt( problem, param )
        print( "logL  = %8.3f  %8.3f  %8.3f" % ( logL, logL0, altL ) )
        assertAAE( logL, altL )
        assertAAE( logL, logL0 )

        scale = 0.1
        param[-1] = scale
        logL = ged.logLikelihood( problem, param )
        altL = ged.logLikelihood_alt( problem, param )
        print( "logL  = %8.3f  %8.3f" % ( logL, altL ) )
        assertAAE( logL, altL )

        scale = 1.0
        param[-1] = scale
        fitIndex = numpy.asarray( [0,1,-1] )
        dL = ged.partialLogL( problem, param, fitIndex )
        nL = ged.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "partalt = ", ged.partialLogL_alt( problem, param, fitIndex ) )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        print( "Using scale = 0.5 and weights." )
        scale = 0.5
        param[-1] = scale
        problem.weights = numpy.asarray( [1,1,2,2], dtype=float )

        logL = ged.logLikelihood( problem, param )

        lLdata = ged.logLdata( problem, param )
        logL0 = numpy.sum( lLdata )
        print( "logL = %8.3f  %8.3f" % ( logL, logL0 ) )
        assertAAE( logL, logL0 )

        dL = ged.partialLogL( problem, param, fitIndex )
        print( "partalt = ", ged.partialLogL_alt( problem, param, fitIndex ) )
        nL = ged.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        scale = 1.0
        for i in range( 11 ) :
            param = numpy.asarray( [0.2*i-1,0,1,1,9,9,1], dtype=float )
            print( param, "  :  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ged.logLikelihood( problem, param ), end="" )
                param[1] += 0.2
            print( "" )

    def test3( self ) :
        print( "====test3 ErrorsInXandYProblem============================" )

        x = numpy.array( [0,1,2,3,4,5,5,6,7,8,9,10], dtype=float )
        y = numpy.array( [1,0,3,2,5,4,6,5,8,7,10,9], dtype=float )

#        x = numpy.append( x, x + 1 )
#        y = numpy.append( y, y + 1 )

#        x = numpy.append( x, x - 0.5 )
#        y = numpy.append( y, y - 0.5 )

        pm = PolynomialModel( 1 )
        ftr = Fitter( x, pm )
        par = ftr.fit( y )
        yfit = pm( x )
        if self.doplot :
            plt.figure( "x-opt", figsize=(6,6) )
            plt.plot( x, y, 'k.' )
            plt.plot( x, yfit, 'r-' )


        mdl = PolynomialModel( 1 )
        mdl.setLimits( -10, 10 )

        ns = NestedSampler( x, mdl, y, problem="errors", limits=[0.1,1] )
        ns.setEngines( ["galilean", "gibbs"] )
        ns.problem.prior = UniformPrior( limits=[-2.0,2.0] )
        #ns.minimumIterations = 1000

#        print( ns.problem.npars )

#        Tools.printclass( ns.problem )
#        Tools.printclass( ns.distribution )
#        ns.end = 4.0
        ns.verbose = 2

        # find the parameters
        #keep = {0:0, 1:1, 3:0.5, 4:9.5, 5:9.5}
        evid = ns.sample( )

        print( "Prob npars :", ns.walkers[0].problem.npars )
        print( "Parameters :", fmt( ns.parameters, max=10 ) )
        print( "StDevs     :", fmt( ns.stdevs, max=10 ) )
        xopt = ns.samples.nuisance
        print( "xdata      :", fmt( x, max=10 ))
        print( "Nuisance   :", fmt( xopt, max=10 ) )
        print( "StdevNuis  :", fmt( ns.samples.stdevNuis, max=10 ))
        print( "Scale      :", fmt( ns.scale ) )


        print( "ftrpars    :", fmt( par, max=10 ) )
        mlap = ns.samples[-1].allpars
        print( "ML pars    :", fmt( mlap[:2], max=10 ) )
        print( "ML Nuis    :", fmt( mlap[2:-1], max=10 ) )
        print( "ML scale   :", fmt( mlap[-1], max=10 ) )

#        assertAAE( mlap[:2], [0,1], 1 )
#        assertAAE( mlap[2:-1], ( x + y ) / 2, 1 )

        print( "ML logL    :", fmt( ns.samples[-1].logL ) )

        print( "ML logL    :", fmt( ns.distribution.logLikelihood( ns.problem, mlap ) ) )
        mlap[:2] = [0.0,1.0]
        print( "ML allp    :", fmt( mlap, max=10 ) )
        print( "ML logL    :", fmt( ns.distribution.logLikelihood( ns.problem, mlap ) ) )

        mlap[2:-1] = ( x + y ) / 2
        print( "ML allp    :", fmt( mlap, max=10 ) )
        print( "ML logL    :", fmt( ns.distribution.logLikelihood( ns.problem, mlap ) ) )

        nd = len( x )

        if not self.doplot :
            return

        yopt = mdl.result( xopt, ns.parameters )
        plt.plot( xopt, yopt, 'k-' )
        for k in range( 0, nd, 1 ):
            plt.plot( [x[k],xopt[k]], [y[k],yopt[k]], 'g-')

        plt.show()


    def test4( self ) :
        print( "====test4 ErrorsInXandYProblem============================" )
        x = numpy.array( [0,1,9,10], dtype=float )
        y = numpy.array( [1,0,10,9], dtype=float )

        mdl = PolynomialModel( 1 )
        mdl.setLimits( -10, 10 )

        ns = NestedSampler( x, mdl, y, problem="errors", limits=[0.1,1] )
#        ns.setEngines( ["galilean", "gibbs"] )
#        ns.setEngines( ["galilean"] )
        ns.problem.prior = UniformPrior( limits=[-2.0,2.0] )
#        ns.verbose=5

        Tools.printclass( ns.problem )
        Tools.printclass( ns.distribution )

        # find the parameters
        keep = {0:0, 1:1, 3:0.5, 4:9.5, 5:9.5}
        keep = {0:0.5, 1:0.9}
        evid = ns.sample( keep=keep )

        if not self.doplot :
            return

        xopt = ns.samples.nuisance
        yopt = mdl.result( xopt, ns.parameters )

        plt.plot( xopt, yopt, 'k-' )
        for k in range( 4 ):
            plt.plot( [x[k],xopt[k]], [y[k],yopt[k]], 'g-')
        plt.show()

    def test5( self ) :
        print( "====test5 MultipleOutputProblem============================" )
        x = numpy.array( [0,1,2,3], dtype=float )
        y = numpy.array( [[1,0],[0,1],[-1,0],[0,-1]], dtype=float )
        w = numpy.array( [0.9,1.0,1.1,1.2], dtype=float )

        mdl = StellarOrbitModel( spherical=False )
        p = [ 0.0, 1.1, 4.0, 0.01, 0.0, 0.0, 0.0]

        problem = MultipleOutputProblem( model=mdl, xdata=x, ydata=y, weights=w )

        print( problem.weights )

        ym = problem.result( p )

        print( y.flatten() )
        print( ym.flatten() )

        res = problem.residuals( p )

        print( res.flatten() )

        dmdx = problem.derivative( p )
        print( dmdx )
        sh = dmdx.shape
        self.assertTrue( sh[0] == 4 and sh[1] == 2 )

        parts = mdl.partial( x, p )
        print( parts[0] )
        sh = parts[1].shape
        self.assertTrue( sh[0] == 4 and sh[1] == 7 )

        partial = problem.partial( p )
        print( partial )
        sh = partial.shape
        self.assertTrue( sh[0] == 8 and sh[1] == 7 )

        a0 = numpy.arange( 6 ).reshape( 3, 2 )
        a1 = numpy.arange( 6 ).reshape( 3, 2 ) + 10
        a2 = numpy.arange( 6 ).reshape( 3, 2 ) + 20

        aa = numpy.append( a0, a1, 1 )
        aa = numpy.append( aa, a2, 1 )

        print( a0 )
        print( aa )
        print( aa.reshape( -1, 2 ) )

    def XXXtest6( self ) :
        print( "====test6 CategoricalProblem============================" )

        x = numpy.array( [[1,4,2,1,3,2,4,2,3,2,1,2,3]], dtype=float ).transpose()
        y = numpy.array( [1,2,2,3,3,2,1,1,2,3,3,2,2] )

#        w = numpy.array( [0.9,1.0,1.1,1.2], dtype=float )

        mdl = DecisionTreeModel( ndim=1, depth=2, kdim=[0,0,0], split=0.5, itypes=[0] )
        p = [ 0.1, 0.5, 0.6, 0.4,
              0.2, 0.3, 0.7, 0.3,
              0.7, 0.2, 0.1, 0.3]
        print( mdl.fullName() )

        problem = CategoricalProblem( model=mdl, xdata=x, ydata=y )

        print( problem.ncateg, problem.npars, fmt( p, max=None ) )
        print( fmt( problem.ydata, max=None ) )
        print( fmt( x, max=None ) )
        print( fmt( mdl.sortXdata( x ), max=None ) )

        lst = [k for k in range( len( p ) )]
        ym = problem.result( p )
        yl = problem.result( lst )
        dy = problem.partial( p )

        print( fmt( ym, max=None ) )
        print( fmt( yl, max=None ) )
        print( fmt( dy, max=None ) )

        res = problem.residuals( p )

        print( fmt( res, max=None ) )

#        dmdx = problem.derivative( p )
#        print( dmdx )

    def test7( self ) :
        print( "====test7 EvidenceProblem============================" )
        nn = 11
        x = numpy.linspace( -1, 1, nn, dtype=float )
        ym = self.x

        model = PolynomialModel( 0 )
        self.assertRaises( ValueError, EvidenceProblem, model=model, xdata=x, ydata=ym )

        model = SplinesDynamicModel( knots=[-1.0, 1.1] )

        problem = EvidenceProblem( model=model, xdata=x, ydata=ym )

        print( problem.shortName() )
        Tools.printclass( problem )

        engs = problem.myEngines()
        self.assertTrue( len( engs ) == 3 )
        self.assertTrue( engs[0] == "birth" )
        self.assertTrue( engs[1] == "death" )
        self.assertTrue( engs[2] == "struct" )

        self.assertTrue( problem.myStartEngine() == "start" )
        self.assertTrue( problem.myDistribution() == "model" )

    def test7a( self ) :
        print( "====test7a EvidenceProblem============================" )
        nn = 11
        x = numpy.linspace( -1, 1, nn, dtype=float )
        ym = self.data

        model = SplinesDynamicModel( knots=[-1.0, 1.0] )
        model.setLimits( lowLimits=-100, highLimits=100 )

        problem = EvidenceProblem( model=model, xdata=x, ydata=ym )
#        problem = ClassicProblem( model=model, xdata=x, ydata=ym )

        ns = NestedSampler( problem=problem )
        ns.verbose = 2
        logE = ns.sample( plot=self.doplot )

        ns = NestedSampler( problem=problem, limits=[0.1,10.0] )
        ns.verbose = 2

        logE = ns.sample( plot=self.doplot )




    def testcr1( self ) :

        print( "===== cyclic results 1 ================" )

        mdl = PolynomialModel( 1 )
        x = numpy.array( [0,1,9,10], dtype=float )
        y = numpy.array( [1,0,10,9], dtype=float )
        problem = Problem( model=mdl, xdata=x, ydata=y )

        Tools.printclass( mdl )

        r0 = numpy.linspace( -0.7, 1.5, 12 )

        r1 = problem.cyclicCorrection( r0 )

        assertAAE( r0, r1 )

    def testcr2( self ) :

        print( "===== cyclic results 2 ================" )

        mdl = PhaseModel( cyclic=1.0 )

        x = numpy.array( [0,1,9,10], dtype=float )
        y = numpy.array( [1,0,10,9], dtype=float )
        problem = Problem( model=mdl, xdata=x, ydata=y )

        Tools.printclass( problem )

        r0 = numpy.linspace( -0.7, 1.5, 12 )
        print( fmt( r0, max=None ) )
        r1 = problem.cyclicCorrection( r0 )
        print( fmt( r1, max=None ) )

        self.assertTrue( all( numpy.abs( r1 ) <= 0.5 ) )

    def testcr3( self ) :

        print( "===== cyclic results 3 ================" )

        mdl = PhaseModel( cyclic=3.0 )

        x = numpy.array( [0,1,9,10], dtype=float )
        y = numpy.array( [1,0,10,9], dtype=float )
        problem = Problem( model=mdl, xdata=x, ydata=y )

        Tools.printclass( problem )

        r0 = numpy.linspace( -4.2, 4.5, 12 ).reshape(3,4).transpose()
        print( fmt( r0, max=None ) )
        r1 = problem.cyclicCorrection( r0 )
        print( fmt( r1, max=None ) )

        self.assertTrue( all( numpy.abs( r1.flatten() ) <= 1.5 ) )

    def testcr4( self ) :

        print( "===== cyclic results 4 ================" )

        mdl = PhaseModel( cyclic={0:1.0} )

        x = numpy.array( [0,1,9,10], dtype=float )
        y = numpy.array( [1,0,10,9], dtype=float )
        problem = Problem( model=mdl, xdata=x, ydata=y )

        Tools.printclass( problem )

        r0 = numpy.linspace( -0.8, 1.5, 12 ).reshape(2,6).transpose()
        print( fmt( r0, max=None ) )
        r1 = problem.cyclicCorrection( r0 )
        print( fmt( r1, max=None ) )

        assertAAE( r0[:,1], r1[:,1] )
        self.assertTrue( all( numpy.abs( r1[:,0] ) < 0.5 ) )

    def testcr5( self ) :

        print( "===== cyclic results 5 ================" )

        mdl = PhaseModel( cyclic={0: 1.0, 2:1.2} )
        x = numpy.array( [0,1,9,10], dtype=float )
        y = numpy.array( [1,0,10,9], dtype=float )
        problem = Problem( model=mdl, xdata=x, ydata=y )

        r0 = numpy.linspace( -0.8, 1.5, 12 ).reshape(3,4).transpose()
        print( fmt( r0, max=None ) )
        r1 = problem.cyclicCorrection( r0 )
        print( fmt( r1, max=None ) )

        assertAAE( r0[:,1], r1[:,1] )
        self.assertTrue( all( numpy.abs( r1[:,0] ) < 0.5 ) )
        self.assertTrue( all( numpy.abs( r1[:,2] ) < 0.6 ) )




    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( ProblemTest.__class__ )


class PhaseModel( PolynomialModel ) :

    def __init__( self, ndim=1, cyclic=1.0, **kwargs ) :

        super().__init__( degree=0, ndim=ndim, **kwargs )

        Tools.setAttribute( self, "cyclic", cyclic )



import numpy as numpy
from astropy import units
import math
from . import Tools
from .NonLinearModel import NonLinearModel

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2015 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class SineDriftModel( NonLinearModel ):
    """
    Sinusoidal Model with drifting frequency.
    .. math::
        f( x:p ) = p_1 * \cos( \phi ) + p_2 * sin( \phi )
        \phi = 2 * \pi * x * ( p_0 + x * p_3 )

    where :math:`p_0` = frequency, :math:`p_3` = is the drift in frequency,
    :math:`p_1` = amplitude cosine and :math:`p_2` = amplitude sine.
    As always x = input.

    The parameters are initialized at {1.0, 1.0, 1.0, 0.0}. It is a non-linear model.

    Examples
    --------
    >>> sine = SineDriftModel( )
    >>> print( sine.npchain )
    >>> pars = [0.1,0,1,0.0]
    >>> sine.parameters = pars
    >>> print( sine( numpy.arange( 101, dtype=float ) ) )     # 10 sine periods, no drift
    >>> pars = [0.1,0,1,0.001]
    >>> sine.parameters = pars
    >>> print( sine( numpy.arange( 101, dtype=float ) ) )     # 10 sine periods, drifting


    Attributes
    ----------
        No attributes of its own.

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    """

    def __init__( self, copy=None, **kwargs ):
        """
        Sinusiodal model with drifting frequency.

        Number of parameters is 4.

        Parameters
        ----------
        copy : SineDriftModel
            to be copied
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.

        """
        param = [1.0, 1.0, 1.0, 0.0]
        names = ["frequency", "cosamp", "sinamp", "freqslope"]

        super( SineDriftModel, self ).__init__( 4, copy=copy, params=param,
                        names=names, **kwargs )

    def copy( self ):
        """ Copy method.  """
        return SineDriftModel( copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        x = 2 * math.pi * xdata * ( params[0] + params[3] * xdata )
        return params[1] * numpy.cos( x ) + params[2] * numpy.sin( x )

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative of f to x (df/dx) at the input values.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        x = 2 * math.pi * xdata * ( params[0] + params[3] * xdata )
        return ( 2 * math.pi * ( params[0] + 2 * params[3] * xdata ) *
                ( params[2] * numpy.cos( x ) - params[1] * numpy.sin( x ) ) )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input values.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        np = self.npbase if parlist is None else len( parlist )
        partial = numpy.ndarray( ( Tools.length( xdata ), np ) )

        x0 = 2 * math.pi * xdata
        x3 = x0 * xdata
        xf = x0 * params[0] + x3 * params[3]
        cx = numpy.cos( xf )
        sx = numpy.sin( xf )
        parts = { 0 : ( lambda: x0 * params[2] * cx - x0 * params[1] * sx ),
                  1 : ( lambda: cx ),
                  2 : ( lambda: sx ),
                  3 : ( lambda: x3 * params[2] * cx - x3 * params[1] * sx ) }


        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return ( "Sine: f( x:p ) = p_1 * cos( 2PI * x * ( p_0 + p_3 * x ) ) + " +
                                  "p_2 * sin( 2PI * x * ( p_0 + p_3 * x ) )" )

    def baseParameterUnit( self, k ):
        """
        Return the name of a parameter.

        Parameters
        ----------
        k : int
            the kth parameter.

        """
        if k == 0:
            return units.Unit( units.si.rad ) / self.xUnit
        if k == 3:
            return units.Unit( units.si.rad ) / ( self.xUnit * self.xUnit )
        return self.yUnit





import numpy as numpy
import math

from .ErrorDistribution import ErrorDistribution
from .LogFactorial import logFactorial
from .Formatter import formatter as fmt
from .Tools import setAttribute as setatt

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Alpha"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2019 - 2022 Do Kester


class DistanceCostFunction( ErrorDistribution ):
    """
    To calculate a distance based cost function

    For one observation with n counts it holds

        f( d ) = exp( -SUM( d / s ) )

    where d are the distances and s is the scale

    The function is mostly used to calculate the likelihood L of
    traveling-salesman-like problems

    Author       Do Kester.

    """

    LOG2 = math.log( 2.0 )
    LOG2P1 = math.log( 2.0 ) + 1

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, copy=None ):
        """
        Constructor.

        Parameters
        ----------
        copy : DistanceCostFunction
            distribution to be copied.

        """
        super( ).__init__( copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return DistanceCostFunction( copy=self )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always false for this distribution.
        """
        return True


    #  *********LIKELIHOODS***************************************************
    def logLikelihood_alt( self, problem, allpars ):
        """
        Return the negative sum of the distances.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem

        """
        self.ncalls += 1

        return -numpy.sum( problem.result( allpars ) )


    def logLdata( self, problem, allpars ):
        """
        Return the individual distances (multiplied by the weights).

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem

        """
        return -problem.result( allpars )

    def partialLogL( self, model, param, fitIndex ):
        """
        Does not work for this class

        Parameters
        ----------
        model : Model
            model to calculate mock data
        param : array_like
            parameters of the model
        fitIndex : array_like
            indices of the params to be fitted
        """
        pass

    def __str__( self ) :
        return "Distance Cost Function"




# run with : python3 -m unittest TestSampleList

import unittest
import numpy as numpy
import sys
from numpy.testing import assert_array_almost_equal as assertAAE
from astropy import units
import math

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2002 Do Kester

class TestWalkerList( unittest.TestCase ):
    """
    Test harness.

    Author       Do Kester

    """

    # Define x independent variable
    x = numpy.asarray( [ -1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0], dtype=float )

    # Define noise: random Gaussian noise sig=0.3
    noise = numpy.asarray( [ -0.000996, -0.046035,  0.013656,  0.418449,  0.0295155,  0.273705,
    -0.204794,  0.275843, -0.415945, -0.373516, -0.158084], dtype=float )
    wgt = numpy.asarray( [1,2,3,4,5,6,7,4,3,2,1], dtype=float )                 #  total 38
    par = numpy.asarray( [3,2,1,0.3], dtype=float )
    fi  = [0,1,2,3]
    len = 11


    #  **************************************************************
    def testWalkerList( self ):
        print( "=========  WalkerListTest  =======================" )
        gm = GaussModel( )
        gm += PolynomialModel( 0 )

        problem = ClassicProblem( gm, xdata=self.x, ydata=self.noise )
#        errdis = GaussErrorDistribution( )

        lnZ = 1.234
        wl0 = WalkerList( problem, self.len, self.par, self.fi )
        self.assertTrue( wl0.logZ == 0 )
        self.assertTrue( wl0.info == 0 )
        self.assertTrue( wl0._count == self.len )
        self.assertTrue( wl0.iteration == 0 )

        k = 0
        for s in wl0 :
            self.assertTrue( s.id == k )
            self.assertTrue( s.parent == -1 )
            self.assertTrue( isinstance( s.problem, ClassicProblem ) )
            self.assertTrue( s.logL == 0 )
            self.assertTrue( s.start == 0 )
            self.assertTrue( s.allpars[0] == 3 )
            self.assertTrue( s.allpars[1] == 2 )
            self.assertTrue( s.allpars[2] == 1 )
            self.assertTrue( s.allpars[3] == 0.3 )

            self.assertTrue( len( s.fitIndex ) == 4 )
            k += 1

        ap = numpy.append( gm.parameters, [0.5] )
        fi = numpy.asarray( [0,1,2,3,-1] )

        wl = WalkerList( problem, self.len, ap, fi )
        k = 0
        for s in wl:
            s.id = k + 1
            s.parent = ( k + 2 ) % self.len + 1
            s.allpars = ap + 0.01 * self.noise[k]
            s.logL = -1213.0 + self.x[k]
            print( s )
            print( "    allpars ", fmt( s.allpars ) )
            print( "    fitindx ", fmt( s.fitIndex ) )
            k += 1
        wl.logZ = lnZ
        wl.info = 30
        self.assertTrue( wl.logZ == lnZ )
        self.assertTrue( wl.info == 30 )


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestWalkerList.__class__ )



if __name__ == '__main__':
    unittest.main()





import numpy as numpy
import math
from . import Tools
from .Tools import setAttribute as setatt
from .LinearModel import LinearModel
from .PolynomialModel import PolynomialModel

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2020 Do Kester

class PolySineAmpModel( LinearModel ):
    """
    Sine of fixed frequency with polynomials as amplitudes.

    Find amplitudes/phases for sinusoidal of a given frequency.

        f( x,y:p ) = P(y) cos( 2 &pi; &omega; x ) + Q(y) sin( 2 &pi; &omega; x )

    P(y), Q(y) are Polynomials of a certain order, n.

    It is a linear model in 2 dimensions, with 2n+2 papameters.

    Examples
    --------
    >>> sine = PolySineAmpModel( 2, 150.0 )        # fixed frequency of 150 Hz

    Attributes
    ----------
    degree : int
        degree of the polynomials
    frequency : float
        frequency of the sine.
    pm : PolynomialModel
        polynomial to be multiplied with the (co)sine

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    """
    def __init__( self, degree, frequency, copy=None, fixed=None, **kwargs ):
        """
        Sine model of a fixed frequency and polynomials as coefficients.

        Number of parameters is 2n+2.

        Parameters
        ----------
        degree : int
            degree of the polynomials
        frequency : float
            the frequency, &omega;
        copy : PolySineAmpModel
            model to be copied
        fixed : dict
            If not None raise AttributeError.

        Raises
        ------
        AttributeError : When fixed is not None

        """
        if fixed is not None :
            raise AttributeError( "PolySineAmpModel cannot have fixed parameters" )

        super( PolySineAmpModel, self ).__init__( 2*degree+2, ndim=2, copy=copy, **kwargs )

        if copy is None :
            self.frequency = frequency
            self.degree = degree
        else :
            self.frequency = copy.frequency
            self.degree = copy.degree

        setatt( self, "pm", PolynomialModel( degree ) )


    def copy( self ):
        """ Copy method.  """
        return PolySineAmpModel( self.degree, self.frequency, copy=self )

    def __setattr__( self, name, value ) :
        if name == "frequency" :
            setatt( self, name, value, type=float )
        elif name == "degree" :
            setatt( self, name, value, type=int )
        elif name == "pm" :
            raise AttributeError( "Attribute pm can not be set" )
        else :
            super( PolySineAmpModel, self ).__setattr__( name, value )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like [2,ndata]
            values at which to calculate the partials
        params : array_like
            parameters of the model (ignored in LinearModels)
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        nxdata = Tools.length( xdata )
        part = numpy.zeros( ( nxdata, self.npbase ), dtype=float )
        x = xdata[:,0]
        y = xdata[:,1]
        cx = numpy.cos( 2 * math.pi * self.frequency * x )
        sx = numpy.sin( 2 * math.pi * self.frequency * x )
        np = self.degree + 1
        for k in range( np ) :
            part[:,k] = cx
            n = k + np
            part[:,n] = sx
            cx *= y
            sx *= y
        return part

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative of f to (x,y) (df/dx,df/dy) at the input value.

        Parameters
        ----------
        xdata : array_like [2,ndata]
            values at which to calculate the partials
        params : array_like
            parameters of the model

        """
        nxdata = Tools.length( xdata )
        dfdx = numpy.zeros( ( nxdata, 2 ), dtype=float )
        x = xdata[:,0]
        y = xdata[:,1]
        np = self.degree + 1
        tpf = 2 * math.pi * self.frequency
        cx = numpy.cos( tpf * x )
        sx = numpy.sin( tpf * x )
        dfdx[:,0] = tpf * ( self.pm.result( y, params[np:] ) * cx -
                            self.pm.result( y, params[:np] ) * sx )
        dfdx[:,1] = ( self.pm.derivative( y, params[:np] ) * cx +
                      self.pm.derivative( y, params[np:] ) * sx )
        return dfdx

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        np = self.npbase // 2
        bn1 = "PolySineAmp: f( x,y:p ) = ( p_0"
        bn2 = "                          ( p_%d"%np
        if np > 1 :
            bn1 += " + p_1 * y"
            bn2 += " + p_%d * y"%(np+1)
        for k in range( 2, np ) :
            bn1 += " + p_%d * y^%d"%(k,k)
            bn2 += " + p_%d * y^%d"%(np+k,np+k)
        bn1 += " ) * cos( 2 pi * x * f ) +\n" + bn2
        bn1 += " ) * sin( 2 pi * x * f ); f = %f"%self.frequency
        return bn1

    def baseParameterUnit( self, k ):
        """
        Return the name of a parameter.
        Parameters
        ----------
        k : int
            the kth parameter.

        """
        np = self.npbase // 2
        return self.yUnit / ( self.xUnit[1] ** ( k % np ) )






from .UniformPrior import UniformPrior

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.6.2"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 - 2020 Do Kester

class CircularUniformPrior( UniformPrior ):
    """
    Circular Uniform prior distribution, for location parameters.
    The lowLimit is wrapped onto the highLimit.

    A wrapper around:
        UniformPrior( circular=... limits=... )

    Examples
    --------
    >>> pr = CircularUniformPrior( circular=math.pi )       # circular between [0,pi]
    >>> pr = CircularUniformPrior( limits=[3,10] )          # circular between [3,10]

    Attributes from Prior
    --------------------=
    lowLimit, highLimit, deltaP, circular, _lowDomain, _highDomain, _umin, _urng

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, circular=None, limits=None, prior=None ):
        """
        Constructor.

        Parameters
        ----------
        limits : array of 2 floats
            [low,high]  range of the prior. Low is wrapped onto high.
        circular : float
            period of circularity
        prior : CircularUniformPrior
            to be copied.

        """
        if circular is None :
            circular = True             ## use limits as period

        super( ).__init__( circular=circular, limits=limits, prior=prior )

    def copy( self ):
        """ Return a (deep) copy of itself. """
        return CircularUniformPrior( prior=self, limits=self.limits, circular=self.circular )

    def __str__( self ):
        """ Return a string representation of the prior.  """
        return str( "CircularUniformPrior " + ( "unbound." if not self.isBound( )
                else ( "period [%.2f..%.2f]"%( self.lowLimit, self.highLimit ) ) ) )






# run with : python3 -m unittest TestAstropyModels.Test.test1

import unittest
import os
import math
import numpy as numpy
from astropy import units
from astropy import modeling
#from astropy.modeling.models import Gaussian1D
#from astropy.modeling.models import Polynomial1D
import matplotlib.pyplot as plt
import warnings

from numpy.testing import assert_array_almost_equal as assertAAE
from numpy.testing import assert_array_equal as assertAE
from StdTests import stdModeltest

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

def ur( x, p ) :
    return p[0] * numpy.sin( p[1] * x + p[2] * numpy.log( x + p[3] ) )


def etalonRes( xdata, params ):
    x = math.pi * xdata * params[2] + params[3]
    sx = numpy.sin( x )
    return params[0] / ( 1.0 + params[1] * sx * sx )


def etalonPart( xdata, params ) :

    partial = numpy.ndarray( ( Tools.length( xdata ), 4 ) )

    x = math.pi * xdata * params[2] + params[3]
    sx = numpy.sin( x )
    s2 = sx * sx
    dd = 1.0 / ( 1 + params[1] * s2 )
    d2 = dd * dd
    p3 = - 2 * params[0] * params[1] * sx * numpy.cos( x ) * d2

    partial[:,0] = dd
    partial[:,1] = -params[0] * s2 * d2
    partial[:,2] = math.pi * xdata * p3
    partial[:,3] = p3

    return partial

def etalonDer( xdata, params ) :

    x = math.pi * xdata * params[2] + params[3]
    sx = numpy.sin( x )
    dd = 1 + params[1] * sx * sx
    dd *= dd
    return - 2 * math.pi * params[0] * params[1] * params[2] * sx * numpy.cos( x ) / dd

def fitfunc(x, p):
    return x[:,0]*p[0] + x[:,1]*p[1] + p[2]


class Test( unittest.TestCase ):
    """
    Test harness for Models

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def test1( self ):
        print( "******USER MODEL 1***********************" )

        m = UserModel( 4, ur, userName="FunkModel" )
        m.parameters = [1, 1, 1, 1]

        p = numpy.asarray( [2.3, 1.1, 3.0, 5.0], dtype=float )

        stdModeltest( m, p, plot=self.doplot )


    def test2( self ):
        print( "******USER MODEL 2***********************" )

        m1 = UserModel( 4, etalonRes, userPartial=etalonPart, userDeriv=etalonDer )
        m1.parameters = [1, 1, 1, 1]

        p = numpy.asarray( [2.3, 1.1, 3.0, 5.0], dtype=float )

        stdModeltest( m1, p, plot=self.doplot )

        N = 201
        x = numpy.arange( N, dtype=float ) / 25 - 2

        m2 = EtalonModel()

        assertAE( m1.result( x, p ), m2.result( x, p ) )        
        assertAE( m1.partial( x, p ), m2.partial( x, p ) )        
        assertAE( m1.derivative( x, p ), m2.derivative( x, p ) )        


    def test3( self ):
        print( "******USER MODEL 2***********************" )

        # Make 10 random 2-dimensional points
        x = numpy.random.rand( 10, 2 )

        # Make fake data + noise
        p = numpy.array( [1,2,3] )
        cleandata = fitfunc( x, p )
        noisedata = cleandata * numpy.random.normal( 1, 0.05, x.shape[0] )

        # Find model parameters from noisy data
        model = UserModel( len(p), fitfunc, ndim=2 )
        fitter = Fitter( x, model )
        param = fitter.fit( noisedata )

        self.assertTrue( model.ndim == 2 )
        print( param )




    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestUserModel.__class__ )

if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
import math
from . import Tools
from .Formatter import formatter as fmt

from .Problem import Problem
from .Sample import Sample

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.6.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2008 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2020 Do Kester

class Walker( object ):
    """
    Walker is member of the cloud of points used in NestedSampler.

    Attributes
    ----------
    id : int
        identification number
    parent : int
        id of the parent (-1 for Adam/Eve)
    start : int
        iteration in which the walker is constructed
    problem : Problem
        the problem being addressed
    logL : float
        log Likelihood = log Prob( data | params )
    allpars : array_like
        list of parameters and hyperparameters
    fitIndex : array_like
        list of (super)parameters to be fitted.
    parameters : array_like (read only)
        parameters (of the model)
    hypars : array_like (read only)
        list of hyper parameters (of the error distribution)

    Author       Do Kester

    """

    def __init__( self, wid, problem, allpars, fitIndex, parent=-1, start=0, copy=None ):
        """
        Constructor.

        Either errdis or copy is obligatory.

        Parameters
        ----------
        wid : int
            id of the walker
        problem : Problem
            the problem being used. Parameters are copied from its model.
        allpars : array_like
            array of parameters and hyperparameters
        fitIndex : None or array_like
            indices of allpars to be fitted
            None is all
        parent : int
            id of the parent (-1 for Adam/Eve)
        start : int
            iteration in which the walker is constructed
        copy : Walker
            the walker to be copied

        """
        self.id = wid
        self.allpars = allpars.copy()
        self.fitIndex = fitIndex

        if copy is None :
            self.start = start
            self.parent = parent
            self.problem = problem
            self.logL = 0.0
        else :
            self.start = copy.start
            self.parent = copy.parent
            self.problem = problem.copy()
            self.logL = copy.logL

    def copy( self ):
        """
        Copy.

        """
        return Walker( self.id, self.problem, self.allpars, self.fitIndex, copy=self )

    def toSample( self, logW ) :
        """
        Return the contents of the Walker as a Sample.
        """

        np = self.problem.npars
        nm = self.problem.model.npars if self.problem.model else np

        param = self.allpars[:nm]
        sample = Sample( self.id, self.parent, self.start, self.problem.model,
                         parameters=param, fitIndex=self.fitIndex )

        if len( self.allpars ) > np :
            sample.hyper = self.allpars[np:]
        if np > nm :
            sample.nuisance = self.allpars[nm:np]

        sample.logL = self.logL
        sample.logW = logW
        return sample

    def __getattr__( self, name ) :
        """
        Return the value of one of `parameters`, `scale`,

        """
        if name == "parameters" :
            np = self.problem.model.npars
            return self.allpars[:np]
        elif name == "hypars" :
            np = self.problem.model.npars
            if len( self.allpars ) > np :
                return self.allpars[np:]
            else :
                return None
        else :
            raise AttributeError( "Unknown attribute " + name )

        return None

    def __setattr__( self, name, value ) :
        """
        Set attributes.
        """
        if name == "allpars" or name == "fitIndex":
            object.__setattr__( self, name, value )
            return

        key1 = {"id" : int, "parent" : int, "start" : int, "problem": Problem,
                "logL" : float }
        if Tools.setSingleAttributes( self, name, value, key1 ) :
            pass
        else :
            raise AttributeError(
                "Object has no attribute " + name + " of type " + str( value.__class__ ) )

    def __str__( self ) :
        return str( "Walker: %3d" % self.id )


    def check( self, nhyp=0, nuisance=0 ) :
        """
        Perform some sanity checks.
        """
        np = self.problem.model.npars
        na = len( self.allpars )
        nm = len( self.problem.model.parameters )

        if not na == ( np + nhyp + nuisance):
            Tools.printclass( self )
            Tools.printclass( self.problem )
            raise ValueError( "Walker %d inconsistent parameter length : %d is not ( %d + %d )" %
                ( self.id, na, np, nhyp ) )

        if not nm == ( na - nhyp - nuisance) :
            Tools.printclass( self )
            Tools.printclass( self.problem.model )
            raise ValueError( "Walker %d inconsistent with model: allpars: %d, model: %d, hyp: %d )" %
                ( self.id, na, nm, nhyp ) )

        ## Does this (all hypars > 0) always have to be true ???
        if nhyp > 0 and self.allpars[-nhyp] <= 0 :
            raise ValueError( "Sample has non-positive hyperparameter: %f" % self.allpars[-nhyp] )



# run with : python3 -m unittest TestSalesmanProblem

import unittest
import os
import numpy as numpy
from astropy import units
from numpy.testing import assert_array_equal as assertAE
from numpy.testing import assert_array_almost_equal as assertAAE
import math
import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import formatter as fmt
from BayesicFitting import fma


__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for solving the SalesmanProblem with NestedSampler.

    Author       Do Kester

    """
    def __init__( self, name ):
        super( ).__init__( name )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def square( self, xdata, pars, roll=1 ) :
        rpar = numpy.roll( pars, -roll )
        return numpy.sum( numpy.square( xdata[pars] - xdata[rpar] ), 1 ) 


    def initProblem( self, np=10, random=False, weights=False, distance="euclid", scale=1 ):

        n2 = np * np
        if random :
            numpy.random.seed( 12345 )
            x0 = numpy.random.rand( n2 ) * np
            x1 = numpy.random.rand( n2 ) * np
        else :
            x0 = numpy.arange( n2, dtype=int ) % np
            x1 = numpy.arange( n2, dtype=int ) // np
        xdata = numpy.append( x0, x1 ).reshape( 2, n2 ).transpose()
        xdata = numpy.array( xdata, dtype=float )

        wgt = numpy.array( [2.0] * n2 ) if weights else None


        problem = SalesmanProblem( xdata, wgt, distance=distance, scale=scale )

#        numpy.set_printoptions( precision=3, suppress=True )
        return problem

    def test1( self ):
        print( "\n   SalesmanProblem Test Manhattan\n" )
        np = 10
        pars = numpy.arange( np*np, dtype=int )

        problem = self.initProblem( np=np, distance="manhattan" )
        print( problem )

        mdis = problem.minimumDistance( )
        print( "mindis  ", mdis )

        dis = problem.result( pars )
        sdis = numpy.sum( dis )
        print( fmt( dis, max=None ), fmt( sdis ) )
        self.assertAlmostEqual( sdis,  198, 6 )

        problem = self.initProblem( np=np, distance="manhattan", scale=3 )
        dis = problem.result( pars )
        sdis = numpy.sum( dis )
        print( fmt( dis, max=None ), fmt( sdis ) )
        self.assertAlmostEqual( sdis,  198/3, 6 )



    def test2( self ):
        print( "\n   SalesmanProblem Test Euclidic\n" )
        np = 10
        pars = numpy.arange( np*np, dtype=int )

        problem = self.initProblem( np=np )
        print( problem )

        mdis = problem.minimumDistance( )
        print( "mindis  ", mdis )

        dis = problem.result( pars )
        sdis = numpy.sum( dis )
        print( fmt( dis, max=None ), fmt( sdis ) )
        print( sdis, ( 90 + 9 * 9.05538514 + 12.72792206 ) )
        self.assertAlmostEqual( sdis, ( 90 + 9 * 9.05538514 + 12.72792206 ), 6 )

        problem = self.initProblem( np=np, weights=True )
        dis = problem.result( pars )
        sdis = numpy.sum( dis )
        print( fmt( dis, max=None ), fmt( sdis ) )
        print( sdis, 2*( 90 + 9 * 9.05538514 + 12.72792206 ) )
        self.assertAlmostEqual( sdis, 2 * ( 90 + 9 * 9.05538514 + 12.72792206 ), 6 )

    def test3( self ):
        print( "\n   SalesmanProblem Test Spherical\n" )
        np = 10
        pars = numpy.arange( np*np, dtype=int )

        problem = self.initProblem( np=np, distance="spheric" )
        problem.xdata[:,0] *= 36
        problem.xdata[:,1] -= 5
        problem.xdata[:,1] *= 18

        print( problem )

        print( problem.xdata.T )

        mdis = problem.minimumDistance( )
        print( "mindis  ", mdis )


        dis = problem.result( pars )
        sdis = numpy.sum( dis )
        print( fmt( dis, max=None ), fmt( sdis ) )
#        print( sdis, ( 90 + 9 * 9.05538514 + 12.72792206 ) )
#        self.assertAlmostEqual( sdis, ( 90 + 9 * 9.05538514 + 12.72792206 ), 6 )

        xd = [[0,0], [90,0], [100,0], [200,0], [270,0], [0,0],  
              [0,10], [0,30], [0,65], [111,90], [111,-10], [111,-90]]

        problem = SalesmanProblem( xd, distance="sheric" )

        print( "scale   ", problem.scale )

        dis = problem.result( pars[:12] ) *180 / math.pi
        
        print( fmt( dis, max=None ) )

        dd = [90, 10, 100, 70, 90, 10, 20, 35, 25, 100, 80, 90]
        assertAAE( dis, dd )

        xx = [[20,33], [43,30], [55,23], [120,33], [143,30], [155,23],
              [220,-33], [243,-30], [255,-23], [120,-33], [143,-30], [155,-23]]

        problem = SalesmanProblem( xx, distance="sheric" )

        dis = problem.result( pars[:12] ) *180 / math.pi
        
        print( fmt( dis, max=None ) )

        assertAAE( dis[0], dis[3] )
        assertAAE( dis[1], dis[4] )
        assertAAE( dis[3], dis[6] )
        assertAAE( dis[4], dis[7] )
        assertAAE( dis[6], dis[9] )
        assertAAE( dis[7], dis[10] )


    def test4( self ):
        print( "\n   SalesmanProblem Test User distance\n" )
        np = 10
        pars = numpy.arange( np*np, dtype=int )

        problem = self.initProblem( np=np, distance=self.square )
        print( problem )

        mdis = problem.minimumDistance( )
        print( "mindis  ", mdis )

        dis = problem.result( pars )
        sdis = numpy.sum( dis )
        print( fmt( dis, max=None ), fmt( sdis ) )
        self.assertAlmostEqual( sdis,  990, 6 )

        problem = self.initProblem( np=np, distance=self.square, scale=3 )
        dis = problem.result( pars )
        sdis = numpy.sum( dis )
        print( fmt( dis, max=None ), fmt( sdis ) )
        self.assertAlmostEqual( sdis,  990/3, 6 )


    def suite( cls ):
        return unittest.TestCase.suite( TestSalesmanProblem.__class__ )


if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
from astropy import units
import re
import warnings
from . import Tools

from .Problem import Problem
from .Formatter import formatter as fmt

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018        Do Kester

class ErrorsInXandYProblem( Problem ):
    """
    A ErrorsInXandYProblem is an optimization of parameters which involves
    the fitting of data to a Model, where both the ydata and the xdata
    contain errors.

    It entails that the xdata are not at the exact locations. They need to
    be optimized too. Consequently the parameters of the model are appended
    with a set of parameters, the length of xdata. These extra parameters
    will contain the target locations of the xdata. The 2-dimensional distance
    between data (xdata,ydata) and (target, model(target)) is minimised.
    The target are nuisance parameters which are not part of the modeling
    solution.

    As the targets need to be optimised they need a Prior. In the present
    implementation there is the same Prior for all targets, which the centered on
    each of the xdata values.
    S.Gull (1989) argues to use a GaussPrior with a scale similar to the errors
    in both X and Y.

    Attributes
    ----------
    prior : Prior
        Priors for the x-axis nuisance parameters.


    Attributes from Problem
    -----------------------
    model, xdata, ydata, weights, partype


    Author :         Do Kester

    """

    #  *************************************************************************
    def __init__( self, model=None, xdata=None, ydata=None, weights=None,
                    prior=None, copy=None ):
        """
        Problem Constructor.

        Parameters
        ----------
        model : Model
            the model to be solved
        xdata : array_like or None
            independent variable
        ydata : array_like or None
            dependent variable
        weights : array_like or None
            weights associated with ydata
        prior : Prior
            prior for the x-axis nuisance parameters. All centered on each of the xdata
        copy : Problem
            to be copied

        """
        super( ).__init__( model=model, xdata=xdata, ydata=ydata, weights=weights, copy=copy )
        self.npars += Tools.length( self.xdata )

        if copy is None :
            self.prior = prior
        else :
            self.prior = copy.prior


    def copy( self ):
        """
        Copy.

        """
        return ErrorsInXandYProblem( copy=self )

    #  *****RESULT**************************************************************
    def result( self, param ):
        """
        Returns the result calculated at the xdatas.

        Parameters
        ----------
        param : array_like
            values for the parameters + nuisance params.

        """
        ( xd, pars ) = self.splitParam( param )
        return self.model.result( xd, pars )


    def splitParam( self, param ) :
        """
        Split the parameters into Model parameters and targets.
        """
        np = self.model.npars
        return ( param[np:], param[:np] )


    def partial( self, param ) :
        """
        Return the partials as a matrix [2*nx,np+nx], where nx is the number of
        datapoints and np the number of parameters in the model.
        The upper left submatrix [nx,np] contains dM/dp
        the upper right submatrix [nx,nx] contains dM/dx on the diagonal
        the lower left submatrix [nx,np] contains zeros
        the lower right submatrix [nx,nx] contains the identity matrix

        """
        ( xd, pars ) = self.splitParam( param )
        np = self.model.npars
        nx = len( xd )
        part = numpy.zeros( ( 2*nx, np+nx ), dtype=float )
        part[:nx,:np] = self.model.partial( xd, pars )
        dfdx = self.model.derivative( xd, pars )
        for k in range( nx ) :
            part[k,k+np] = dfdx[k]
            part[k+nx,k+np] = 1.0
        return part

    def derivative( self, param ) :
        """
        Return the derivative to the Model.

        Parameters
        ----------
        params : array_like
            list of problem parameters
        """

        ( xd, pars ) = self.splitParam( param )
        return self.model.derivative( xd, pars )

    def domain2Unit( self, dval, kpar ) :
        """
        Return value in [0,1] for the selected parameter.

        Parameters
        ----------
        dval : float
            domain value for the selected parameter
        kpar : int
            selected parameter index, where kp is index in [parameters, hyperparams]
        """
        if kpar < self.model.npars :
            return self.model.domain2Unit( dval, kpar )

        kp = kpar - self.model.npars
        dv = dval - self.xdata[kp]

        return self.prior.domain2Unit( dv )

    def unit2Domain( self, uval, kpar ) :
        """
        Return domain value for the selected parameter.

        Parameters
        ----------
        uval : array_like
            unit value for the selected parameter
        kpar : int
            selected parameter indices, where kp is index in [parameters, hyperparams]
        """
        if kpar < self.model.npars :
            return self.model.unit2Domain( uval, kpar )

        return self.prior.unit2Domain( uval ) + self.xdata[kpar-self.model.npars]

    def residuals( self, param, mockdata=None ) :
        """
        Return the (weighted) true distance between (xdata,ydata) and (xtry,ytry) where xtry are
        the trial values for xdata and ytry = model.result( xtry, param )

        res = hypothenusa( xdata - xtry, ydata - ytry ) * weights

        Parameters
        ----------
        param : array_like
            model parameters and xdata parameters
        mockdata : array_like
            model fit data model.result( xtry, param )
        """
        ( xd, pars ) = self.splitParam( param )
        res = self.ydata - self.result( param )
        return numpy.hypot( res, self.xdata - xd )


    def weightedResiduals( self, param, mockdata=None, extra=False ) :
        """
        Returns the (weighted) residuals, calculated at the xdata.

        Parameters
        ----------
        param : array_like
            values for the parameters.
        mockdata : array_like
            model fit at xdata
        extra : bool (False)
            true  : return ( wgt * dist, wgt * [yres,xres] / dist )
            false : return wgt * dist
        """
        ( xd, pars ) = self.splitParam( param )
        yres = self.ydata - self.result( param )
        xres = self.xdata - xd
        dist = numpy.hypot( xres, yres )

        if extra :
            if self.weights is None :
                exres = numpy.append( yres / dist, xres / dist )
                return ( dist, exres )
            else :
                exres = numpy.append( self.weights * yres / dist, self.weights * xres / dist )
                return ( dist * self.weights, exres )
        else :
            return dist if self.weights is None else dist * self.weights



    def weightedResSq( self, param, mockdata=None, extra=False ) :
        """
        Return the (weighted) squared distance between (xdata,ydata) and (xtry,ytry) where xtry are
        the trial values for xdata and ytry = model.result( xtry, param )

        Parameters
        ----------
        param : array_like
            model parameters and xdata parameters
        mockdata : array_like
            model fit data model.result( xtry, param )
        extra : bool (False)
            true  : return ( wgt * res^2, wgt * [yres,xres] )
            false : return wgt * res^2
        """
        ( xd, pars ) = self.splitParam( param )
        yres = self.ydata - self.result( param )
        xres = self.xdata - xd

        res2 = numpy.square( yres ) + numpy.square( xres )
        if extra :
            if self.weights is not None :
                res2 *= self.weights
                xres *= self.weights
                yres *= self.weights
            return ( res2, numpy.append( yres, xres ) )

        return res2 if self.weights is None else res2 * self.weights

    def myEngines( self ) :
        """
        Return a default list of preferred engines
        """
        return ["galilean", "gibbs", "chord"]

    def myStartEngine( self ) :
        """
        Return a default preferred start engines: "start"
        """
        return "start"

    def myDistribution( self ) :
        """
        Return a default preferred ErrorDistribution: "gauss"
        """
        return "gauss"

    #  *****TOSTRING***********************************************************
    def __str__( self ):
        """ Returns a string representation of the model.  """
        return "ErrorsInXandYProblem of %s" % self.model






import math
from scipy import special
from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017 - 2021 Do Kester


class GaussPrior( Prior ):
    """
    Gauss prior distribution.

        Pr( x ) = exp( - ( ( x - c ) / scale )^2 )

    By default: center = 0 and scale = 1.

    It can also have a limited domain. (To be done)
    By default the domain is [-Inf,+Inf].
    In computational practice the domain is limited to about [-6, 6] scale units.

    domain2unit: u = 0.5 * ( erf( ( d - center ) / scale ) + 1 )
    unit2domain: d = erfinv( 2 * u - 1 ) * scale + center

    Examples
    --------
    >>> pr = GaussPrior()                         # center=0, scale=1
    >>> pr = GaussPrior( center=1.0, scale=0.5 )
    >>> pr = GaussPrior( limits=[0,None] )        # limited to values >= 0
    >>> pr = GaussPrior( center=1, circular=3 )   # circular between 0.5 and 2.5

    Attributes
    ----------
    center : float
        center of the Gaussian prior
    scale : float
        scale of the Gaussian prior

    Attributes from Prior
    --------------------=
    lowLimit, highLimit, circular, deltaP, _lowDomain, _highDomain


    """

    SPI = 1.0 / math.sqrt( math.pi )
    LSPI = math.log( SPI )

    MAXVAL = 6

    #  *********CONSTRUCTOR***************************************************
    def __init__( self, center=0.0, scale=1.0, limits=None, circular=False, prior=None ):
        """
        Constructor.

        Parameters
        ----------
        center : float
            of the location of the prior
        scale : float
            of the exponential
        limits : None or [float,float]
            None    no limits are set
            2 floats    lowlimit and highlimit
        circular : bool or float
            bool : y|n circular with period from limits[0] to limits[1]
            float : period of circularity
        prior : GaussPrior
            prior to copy (with new scale if applicable)

        """
        self.center = center
        self.scale = scale

        super( ).__init__( limits=limits, circular=circular, prior=prior )

    def copy( self ):
        """ Copy the prior """
        return GaussPrior( prior=self, center=self.center, scale=self.scale,
                           limits=self.limits, circular=self.circular )

    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for a Gauss distribution.

        domain2unit: u = 0.5 * ( erf( ( d - center ) / scale ) + 1 )

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        return 0.5 * ( special.erf( ( dval - self.center ) / self.scale ) + 1 )

    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for a Gauss distribution.

        unit2domain: d = erfinv( 2 * u - 1 ) * scale + center

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        dom = special.erfinv( 2 * uval - 1 )
        if math.isfinite( dom ) :
            return dom * self.scale + self.center
        else :
            fs = - self.MAXVAL if uval < 0.5 else self.MAXVAL
            return self.center + fs * self.scale

#    def partialDomain2Unit( self, dval ):
    def result( self, x ):
        """
        Return a the result of the distribution function at x.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        xs = ( x - self.center ) / self.scale
        return self.SPI * math.exp( - xs * xs )

    def logResult( self, x ):
        """
        Return a the log of the result of the prior.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        xs = ( x - self.center ) / self.scale
        return self.LSPI - ( xs * xs )

    def partialLog( self, x ):
        """
        Return partial derivative of log( Prior ) wrt x.

        Parameters
        ----------
        x : float
            the value

        """
        return -2 * ( x - self.center ) / ( self.scale * self.scale )

    def isBound( self ):
        """ Return true if the integral over the prior is bound.  """
        return True

    def shortName( self ):
        """ Return a string representation of the prior.  """
        return str( "GaussPrior" )



